# Copyright 2022 Cisco Systems, Inc. and its affiliates

"""Methods covering essential API endpoints and related data classes."""
from __future__ import annotations

import logging
from contextlib import contextmanager
from typing import TYPE_CHECKING, Iterator, List, Union

from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from catalystwan.dataclasses import BfdSessionData, Connection, Device, WanInterface
from catalystwan.endpoints.real_time_monitoring.reboot_history import RebootEntry
from catalystwan.exceptions import CatalystwanException
from catalystwan.typed_list import DataSequence
from catalystwan.utils.creation_tools import create_dataclass
from catalystwan.utils.operation_status import OperationStatus
from catalystwan.utils.personality import Personality
from catalystwan.utils.reachability import Reachability

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


logger = logging.getLogger(__name__)


class DevicesAPI:
    """API methods of vManage for getting devices and controllers.

    Attributes:
        session: logged in API client session
    """

    max_params = 1000

    def __init__(self, session: ManagerSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    @property
    def system_ips(self) -> List[str]:
        """List of device system IP addresses."""
        return [device.local_system_ip for device in self.get()]

    @property
    def ips(self):
        """List of device IP addresses."""
        return [device.id for device in self.get()]

    def get_system_ip_based_on_local_system_ip(self, local_system_ip) -> str:
        for dev in self.get():
            if local_system_ip == dev.local_system_ip:
                return dev.id
        return ""

    def get_device_details(self, uuid: str) -> Device:
        """Gets system information for a device.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            Device object
        """
        response = self.session.get(f"/dataservice/system/device/vedges?uuid={uuid}")

        devices = response.dataseq(Device)
        assert len(devices) == 1, "Expected system info response list to have one member"

        return devices[0]

    def count_devices(self, personality: Personality) -> int:
        """Gets number of devices of given personality.

        Args:
            personality: personality of the device

        Returns:
            count of devices
        """
        return sum([1 for device in self.get() if device.personality == personality])

    def get_reachable_devices(self, personality: Personality) -> DataSequence[Device]:
        """Get reachable devices by personality.

        Args:
            personality: personality of the device

        Returns:
            reachable devices
        """
        unsupported_personality = [Personality.VMANAGE]
        assert personality not in unsupported_personality, "Unsupported personality for reachable endpoint"

        devices = self.session.get(f"/dataservice/device/reachable?personality={personality.value}")
        return devices.dataseq(Device)

    def send_certificate_state_to_controllers(
        self,
        sleep_seconds: int = 5,
        timeout_seconds: int = 600,
    ) -> bool:
        """Sending the current status of certificates to the controllers

        Returns:
            bool: True if all ok, False like something wrong
        """

        def _log_exception(retry_state):
            self.session.__get_logger(f"Orignial exception: {retry_state.outcome.exception()}.")

        def check_state(action_data):
            list_action = [action["status"] == OperationStatus.SUCCESS.value for action in action_data]
            return not all(list_action)

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_state),
            retry_error_callback=_log_exception,
        )
        def wait_for_state():
            status_api = f"/dataservice/device/action/status/{action_id}"
            return self.session.get_data(f"{status_api}")

        response = self.session.post("/dataservice/certificate/vedge/list?action=push").json()
        if response.get("id"):
            action_id = response["id"]
        else:
            raise CatalystwanException("Failed to push edges list certificates")

        return True if wait_for_state() else False

    def get(self, rediscover: bool = False) -> DataSequence[Device]:
        """Data sequence of all devices.

        Args:
            rediscover: Rediscover device request payload

        Returns:
            DataSequence[Device] of all devices

        ## Examples:

        Get all vManages:
        >>> devices = DevicesAPI(session).get()
        >>> vManages = devices.filter(personality=Personality.VMANAGE)
        """
        if rediscover:
            logger.info("Rediscovering devices...")
            api = "/dataservice/device/action/rediscoverall"
            self.session.post(url=api)
        devices = self.session.endpoints.monitoring_device_details.list_all_devices()
        device_ids = [device.device_id for device in devices]
        devices_sys_info = DataSequence(Device, [])
        for i in range(0, len(device_ids), self.max_params):
            params = {"deviceId": device_ids[i : i + self.max_params]}
            resp = self.session.get(url="/dataservice/device/system/info", params=params)
            devices_sys_info += resp.dataseq(Device)

        return devices_sys_info


class DeviceStateAPI:
    """Basic API methods of vManage.

    Attributes:
        session: logged in API client session
    """

    def __init__(self, session: ManagerSession) -> None:
        self.session = session

    def __str__(self) -> str:
        return str(self.session)

    def get_device_crash_info(self, device_id: str) -> Union[list, dict]:
        """Gets crash info for a device.

        Args:
             device_id: device ID (usually system-ip)

        Returns:
            Union[list, dict]: list of dicts (FIXME: add mapping to a NamedTuples)
        """
        return self.session.get_data(f"/dataservice/device/crashlog?deviceId={device_id}")

    def get_device_control_connections_info(self, device_id) -> List[Connection]:
        """Gets control connections for a device.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            list of Connection objects
        """
        items = self.session.get_data(f"/dataservice/device/control/connections?deviceId={device_id}")

        return [create_dataclass(Connection, item) for item in items]

    def get_device_orchestrator_connections_info(self, device_id) -> List[Connection]:
        """Gets orchestrator connections for a device

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            list of Connection objects
        """
        items = self.session.get_data(f"/dataservice/device/orchestrator/connections?deviceId={device_id}")

        return [create_dataclass(Connection, item) for item in items]

    def get_device_reboot_history(self, device_id) -> DataSequence[RebootEntry]:
        """Gets device reboots list.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
            list of Reboot objects
        """
        params = {"deviceId": device_id}
        return self.session.endpoints.real_time_monitoring.reboot_history.create_reboot_history_list(params)

    def get_system_status(self, device_id: str) -> Device:
        """Get system information for a device.

        Args:
            device_id: device ID (usually system-ip)

        Returns:
           Device object
        """
        devices = self.session.get_data(f"/dataservice/device/system/info?deviceId={device_id}")

        assert len(devices) == 1, "Expected system info response list to have one member"

        return create_dataclass(Device, devices[0])

    def get_device_wan_interfaces(self, device_id: str):
        wan_interfaces = self.session.get_data(f"/dataservice/device/control/waninterface?deviceId={device_id}")
        return [create_dataclass(WanInterface, wan_ifc) for wan_ifc in wan_interfaces]

    def get_colors(self, device_id: str) -> List[str]:
        url = "/dataservice/device/bfd/state/device/tlocInterfaceMap"
        colors_raw = DevicesAPI(self.session).session.get_json(url + f"?deviceId={device_id}")
        colors = list(colors_raw["intfList"].keys())

        return colors

    @contextmanager
    def enable_data_stream(self) -> Iterator:  # TODO check
        try:
            url_path = "/dataservice/settings/configuration/vmanagedatastream"
            data_stream_status = self.session.get_data(url_path)[0]
            query = {  # TODO Dict[str, obj]
                "enable": True,
                "ipType": "systemIp",
                "serverHostName": "systemIp",
                "vpn": "0",
            }
            url_path = "/dataservice/settings/configuration/vmanagedatastream"
            self.session.post(url=url_path, json=query)
            yield None
        finally:
            url_path = "/dataservice/settings/configuration/vmanagedatastream"
            self.session.post(url=url_path, json=data_stream_status)

    def get_bfd_sessions(self, device_id: str) -> List[BfdSessionData]:
        items = self.session.get_data(f"/dataservice/device/bfd/sessions?deviceId={device_id}")

        return [create_dataclass(BfdSessionData, item) for item in items]

    def wait_for_bfd_session_up(
        self,
        system_ip: str,
        sleep_seconds: int = 5,
        timeout_seconds: int = 60,
        exp_state: str = "up",
    ):
        def check_state(bfd_sessions):
            return not all([bfd_session.state == exp_state for bfd_session in bfd_sessions])

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_state),
        )
        def wait_for_bfd_session_come_up():
            return self.get_bfd_sessions(system_ip)
            # from my observation it is necessary to wait minimum 5 seconds for BFD's session gets up state

        wait_for_bfd_session_come_up()

    def wait_for_device_state(
        self,
        device_id: str,
        sleep_seconds: int = 5,
        timeout_seconds: int = 600,
        exp_state: Reachability = Reachability.REACHABLE,
    ):
        """
        Waiting for the state of the machine.

        Args:
          device_id(Str): Device ID (usually system-ip)
          timeout_seconds(int): Failure timeout.
          sleep_seconds(int): Sleep time.
          exp_state(Reachability): The expected state of the machine

        Returns:
          True if the expected state has been achieved

        """

        def _log_exception(retry_state):
            self.session.__get_logger(f"Orignial exception: {retry_state.outcome.exception()}.")

        def check_state(state):
            return state == exp_state.value

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_state),
            retry_error_callback=_log_exception,
        )
        def wait_for_state():
            return self.get_system_status(device_id).reachability

        return True if wait_for_state() else False


__all__ = ["Device"]



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable

CountryCode = Literal[
    "AE",
    "AR",
    "AT",
    "AU",
    "BA",
    "BB",
    "BE",
    "BG",
    "BH",
    "BN",
    "BO",
    "BR",
    "BY",
    "BA",
    "CA2",
    "CH",
    "CL",
    "CM",
    "CN",
    "CO",
    "CR",
    "CY",
    "CZ",
    "DE",
    "DK",
    "DO",
    "DZ",
    "EC",
    "EE",
    "EG",
    "ES",
    "FI",
    "FJ",
    "FR",
    "GB",
    "GH",
    "GI",
    "GR",
    "HK",
    "HR",
    "HU",
    "ID",
    "IE",
    "IL",
    "IO",
    "IN",
    "IQ",
    "IS",
    "IT",
    "J2",
    "J4",
    "JM",
    "JO",
    "KE",
    "KN",
    "KW",
    "KZ",
    "LB",
    "LI",
    "LK",
    "LT",
    "LU",
    "LV",
    "LY",
    "MA",
    "MC",
    "ME",
    "MK",
    "MN",
    "MO",
    "MT",
    "MX",
    "MY",
    "NL",
    "NO",
    "NZ",
    "OM",
    "PA",
    "PE",
    "PH",
    "PH2",
    "PK",
    "PL",
    "PR",
    "PT",
    "PY",
    "QA",
    "RO",
    "RS",
    "RU",
    "SA",
    "SE",
    "SG",
    "SI",
    "SK",
    "TH",
    "TN",
    "TR",
    "TW",
    "UA",
    "US",
    "UY",
    "VE",
    "VN",
    "ZA",
]

RadioType = Literal[
    "all",
    "24ghz",
    "5ghz",
]

QosProfile = Literal[
    "platinum",
    "gold",
    "silver",
    "bronze",
]

SecurityType = Literal[
    "open",
    "personal",
    "enterprise",
]


class MeStaticIpConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    me_ipv4_address: Union[Global[str], Variable]
    netmask: Union[Global[str], Variable]
    default_gateway: Union[Global[str], Variable] = Field(
        serialization_alias="defaultGateway", validation_alias="defaultGateway"
    )


class MeIpConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    me_dynamic_ip_enabled: Union[Global[bool], Default[bool]] = Field(
        serialization_alias="meDynamicIpEnabled",
        validation_alias="meDynamicIpEnabled",
        default=Default[bool](value=True),
    )
    me_static_ip_config: Optional[MeStaticIpConfig] = None


class SecurityConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    security_type: Global[SecurityType] = Field(serialization_alias="securityType", validation_alias="securityType")
    radius_server_ip: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="radiusServerIp", validation_alias="radiusServerIp", default=None
    )
    radius_server_port: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="radiusServerPort", validation_alias="radiusServerPort", default=None
    )
    radius_server_secret: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="radiusServerSecret", validation_alias="radiusServerSecret", default=None
    )
    passphrase: Optional[Union[Global[str], Variable]] = None


class SSID(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Global[str]
    admin_state: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="adminState", validation_alias="adminState", default=Default[bool](value=True)
    )
    broadcast_ssid: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="broadcastSsid", validation_alias="broadcastSsid", default=Default[bool](value=True)
    )
    vlan_id: Union[Global[int], Variable] = Field(serialization_alias="vlanId", validation_alias="vlanId")
    radio_type: Union[Global[RadioType], Variable, Default[RadioType]] = Field(
        serialization_alias="radioType", validation_alias="radioType", default=Default[RadioType](value="all")
    )
    security_config: SecurityConfig = Field(serialization_alias="securityConfig", validation_alias="securityConfig")
    qos_profile: Union[Global[QosProfile], Variable, Default[QosProfile]] = Field(
        serialization_alias="qosProfile",
        validation_alias="qosProfile",
        default=Default[QosProfile](value="silver"),
    )


class WirelessLanData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    enable_2_4G: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="enable24G", validation_alias="enable24G", default=Default[bool](value=True)
    )
    enable_5G: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="enable5G", validation_alias="enable5G", default=Default[bool](value=True)
    )
    ssid: List[SSID]
    country: Union[Global[CountryCode], Variable]
    username: Union[Global[str], Variable]
    password: Union[Global[str], Variable]
    me_ip_config: MeIpConfig = Field(
        serialization_alias="meIpConfig", validation_alias="meIpConfig", default=MeIpConfig()
    )


class WirelessLanCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: WirelessLanData
    metadata: Optional[dict] = None



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"


from catalystwan.endpoints import APIEndpoints, delete, get, post, put
from catalystwan.models.url_monitoring import DeleteUrlMonitorRequest, UrlMonitoringConfig, UrlMonitoringInfo
from catalystwan.typed_list import DataSequence


class UrlMonitoring(APIEndpoints):
    @get("/url/monitor")
    def get_url_monitor(self) -> DataSequence[UrlMonitoringInfo]:
        ...

    @post("/url/monitor")
    def add_url_monitor(self, payload: UrlMonitoringConfig) -> None:
        ...

    @put("/url/monitor")
    def update_url_monitor(self, payload: UrlMonitoringConfig) -> None:
        ...

    @delete("/url/monitor")
    def delete_url_monitor(self, params: DeleteUrlMonitorRequest) -> None:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

import json
import os
import unittest
from typing import Any, List, cast

from catalystwan.session import create_manager_session
from catalystwan.utils.feature_template.find_template_values import find_template_values


class TestFindTemplateValues(unittest.TestCase):
    def setUp(self) -> None:
        self.session = create_manager_session(
            url=cast(str, os.environ.get("TEST_VMANAGE_URL")),
            port=cast(int, int(os.environ.get("TEST_VMANAGE_PORT"))),  # type: ignore
            username=cast(str, os.environ.get("TEST_VMANAGE_USERNAME")),
            password=cast(str, os.environ.get("TEST_VMANAGE_PASSWORD")),
        )
        self.templates = self.session.api.templates._get_feature_templates(summary=False)

    def test_find_template_value(self):
        for template in self.templates:
            definition = json.loads(template.template_definiton)
            with self.subTest(template_name=template.name):
                parsed_values = find_template_values(definition)
                self.assertFalse(
                    self.is_key_present(parsed_values, ["vipType", "vipValue", "vipVariableName", "vipObjectType"])
                )

    def is_key_present(self, d: dict, keys: List[Any]):
        """
        Checks if any key from keys is present within the dictionary d
        """
        for key, value in d.items():
            if key in keys:
                return True
            if isinstance(value, dict):
                if self.is_key_present(value, keys):
                    return True
            if isinstance(value, list):
                for v in value:
                    if isinstance(v, dict):
                        if self.is_key_present(v, keys):
                            return True
        return False

    def tearDown(self) -> None:
        self.session.close()



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"

from catalystwan.endpoints import APIEndpoints, get
from catalystwan.models.monitoring.security_policy import SecurityPolicyDeviceList


class MonitoringSecurityPolicy(APIEndpoints):
    @get("/security/policy/devicelist", "data")
    def get_device_list(self) -> SecurityPolicyDeviceList:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable

VirtualApplicationType = Literal["dreopt"]

ResourceProfile = Literal[
    "small",
    "medium",
    "large",
    "extra-large",
    "default",
]

AppqoeDeviceRole = Literal[
    "forwarder",
    "forwarderAndServiceNode",
    "serviceNode",
    "serviceNodeWithDre",
    "forwarderAndServiceNodeWithDre",
]

AppnavControllerGroupName = Literal["ACG-APPQOE"]
ServiceNodeGroupName = Literal["SNG-APPQOE"]
ForwarderAndServiceNodeAddress = Literal["192.168.2.2"]  # TODO: 1.Is it really constant? 2.Use ipaddress.IPv4Address?
ForwarderAndServiceNodeControllerAddress = Literal[
    "192.168.2.1"
]  # TODO: 1.Is it really constant? 2.Use ipaddress.IPv4Address?
ServiceNodeExternalAddress = Literal["192.168.2.2"]  # TODO: 1.Is it really constant? 2.Use ipaddress.IPv4Address?
ServiceNodeExternalVpgIp = Literal["192.168.2.1/24"]  # TODO: 1.Is it really constant? 2.Use ipaddress.IPv4Address?


class VirtualApplication(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    instance_id: Global[int] = Field(
        default=Global(value=1), serialization_alias="instanceId", validation_alias="instanceId"
    )
    application_type: Global[VirtualApplicationType] = Field(
        default=Global[VirtualApplicationType](value="dreopt"),
        serialization_alias="applicationType",
        validation_alias="applicationType",
    )
    resource_profile: Union[Global[ResourceProfile], Default[str]] = Field(
        default=Global[ResourceProfile](value="default"),
        serialization_alias="resourceProfile",
        validation_alias="resourceProfile",
    )


class Appqoe(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Default[str] = Default(value="/1")
    appnav_controller_group: Global[AppnavControllerGroupName] = Field(
        default=Global[AppnavControllerGroupName](value="ACG-APPQOE"),
        serialization_alias="appnavControllerGroup",
        validation_alias="appnavControllerGroup",
    )
    service_node_group: Global[ServiceNodeGroupName] = Field(
        default=Global[ServiceNodeGroupName](value="SNG-APPQOE"),
        serialization_alias="serviceNodeGroup",
        validation_alias="serviceNodeGroup",
    )
    service_node_groups: List[Global[ServiceNodeGroupName]] = Field(
        default=[Global[ServiceNodeGroupName](value="SNG-APPQOE")],
        serialization_alias="serviceNodeGroups",
        validation_alias="serviceNodeGroups",
    )
    enable: Global[bool] = Global[bool](value=True)
    vpn: Union[Global[int], Default[None], Variable] = Field(default=Global[int](value=0))


class ServiceContext(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    appqoe: List[Appqoe]


# Frowarder


class ServiceNodeInformation(BaseModel):
    address: Global[str]


class ForwarderController(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Global[str], Variable]
    vpn: Global[int] = Global[int](value=1)


class ForwarderAppnavControllerGroup(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_name: Default[AppnavControllerGroupName] = Field(
        default=Default(value="ACG-APPQOE"), serialization_alias="groupName", validation_alias="groupName"
    )
    appnav_controllers: List[ForwarderController] = Field(
        serialization_alias="appnavControllers", validation_alias="appnavControllers"
    )


class ForwarderNodeGroup(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Union[Global[str], Default[ServiceNodeGroupName]]
    internal: Default[bool] = Default[bool](value=False)
    service_node: List[ServiceNodeInformation] = Field(
        serialization_alias="serviceNode", validation_alias="serviceNode"
    )


class ForwarderRole(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    appnav_controller_group: List[ForwarderAppnavControllerGroup] = Field(
        serialization_alias="appnavControllerGroup", validation_alias="appnavControllerGroup"
    )
    service_node_group: List[ForwarderNodeGroup] = Field(
        serialization_alias="serviceNodeGroup", validation_alias="serviceNodeGroup"
    )
    service_context: ServiceContext = Field(serialization_alias="serviceContext", validation_alias="serviceContext")


# Forwarder and Service


class ServiceNodeInformationDefault(BaseModel):
    address: Default[ForwarderAndServiceNodeAddress] = Default[ForwarderAndServiceNodeAddress](value="192.168.2.2")


class ForwarderAndServiceNodeController(BaseModel):
    address: Default[ForwarderAndServiceNodeControllerAddress] = Default[ForwarderAndServiceNodeControllerAddress](
        value="192.168.2.1"
    )


class ForwarderAndServiceNodeAppnavControllerGroup(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    group_name: Default[AppnavControllerGroupName] = Field(
        default=Default[AppnavControllerGroupName](value="ACG-APPQOE"),
        serialization_alias="groupName",
        validation_alias="groupName",
    )
    appnav_controllers: List[ForwarderAndServiceNodeController] = Field(
        serialization_alias="appnavControllers", validation_alias="appnavControllers"
    )


class ForwarderAndServiceNodeGroup(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Default[ServiceNodeGroupName] = Default[ServiceNodeGroupName](value="SNG-APPQOE")
    internal: Default[bool] = Default[bool](value=True)
    service_node: List[ServiceNodeInformationDefault] = Field(
        serialization_alias="serviceNode", validation_alias="serviceNode"
    )


class ForwarderAndServiceNodeRole(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    appnav_controller_group: List[ForwarderAndServiceNodeAppnavControllerGroup] = Field(
        serialization_alias="appnavControllerGroup", validation_alias="appnavControllerGroup"
    )
    service_node_group: List[ForwarderAndServiceNodeGroup] = Field(
        serialization_alias="serviceNodeGroup", validation_alias="serviceNodeGroup"
    )

    service_context: ServiceContext = Field(serialization_alias="serviceContext", validation_alias="serviceContext")


# Service


class ServiceNodeInformationExternal(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Default[ServiceNodeExternalAddress] = Default[ServiceNodeExternalAddress](value="192.168.2.2")
    vpg_ip: Default[ServiceNodeExternalVpgIp] = Field(
        default=Default[ServiceNodeExternalVpgIp](value="192.168.2.1"),
        serialization_alias="vpgIp",
        validation_alias="vpgIp",
    )


class ServiceNodeGroup(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Default[ServiceNodeGroupName] = Default[ServiceNodeGroupName](value="SNG-APPQOE")
    external_node: Default[bool] = Field(
        default=Default[bool](value=True), serialization_alias="externalNode", validation_alias="externalNode"
    )
    service_node: List[ServiceNodeInformationExternal] = Field(
        default=[ServiceNodeInformationExternal()], serialization_alias="serviceNode", validation_alias="serviceNode"
    )


class ServiceNodeRole(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    service_node_group: List[ServiceNodeGroup] = Field(
        default=[ServiceNodeGroup()], serialization_alias="serviceNodeGroup", validation_alias="serviceNodeGroup"
    )


class AppqoeData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dreopt: Optional[Union[Global[bool], Default[bool]]] = Default[bool](value=False)
    virtual_application: Optional[List[VirtualApplication]] = Field(
        serialization_alias="virtualApplication", validation_alias="virtualApplication"
    )
    appqoe_device_role: Global[str] = Field(
        default=Global(value="forwarder"), serialization_alias="appqoeDeviceRole", validation_alias="appqoeDeviceRole"
    )

    forwarder: Optional[ForwarderRole]
    forwarder_and_service_node: Optional[ForwarderAndServiceNodeRole] = Field(
        serialization_alias="forwarderAndServiceNode", validation_alias="forwarderAndServiceNode"
    )
    service_node: Optional[ServiceNodeRole] = Field(serialization_alias="serviceNode", validation_alias="serviceNode")


class AppqoeCreationPayload(BaseModel):
    name: str
    description: Optional[str] = None
    data: AppqoeData
    metadata: Optional[dict] = None



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Optional

from pydantic import BaseModel, Field


class UrlMonitoringInfo(BaseModel):
    url: Optional[str] = None
    threshold: Optional[int] = None
    alarm_raised: Optional[bool] = Field(
        default=None, serialization_alias="lastUpdatedBy", validation_alias="lastUpdatedBy"
    )


class UrlMonitoringConfig(BaseModel):
    url: str = Field(..., examples=["/client/server/ready"], description="URL registered for monitoring requests.")
    threshold: int = Field(
        ge=10,
        le=100,
        title="Threshold",
        description="vManage alarm is raised after reaching the threshold. "
        "Threshold should be within a range of 10 and 100 inclusive.",
    )


class DeleteUrlMonitorRequest(BaseModel):
    url: str = Field(..., examples=["/client/server/ready"], description="URL registered for monitoring requests.")



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import json
import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, Type, overload

from ciscoconfparse import CiscoConfParse  # type: ignore

from catalystwan.api.task_status_api import Task
from catalystwan.api.templates.cli_template import CLITemplate
from catalystwan.api.templates.device_template.device_template import (
    DeviceSpecificValue,
    DeviceTemplate,
    GeneralTemplate,
)
from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.api.templates.feature_template_field import FeatureTemplateField
from catalystwan.api.templates.feature_template_payload import FeatureTemplatePayload
from catalystwan.api.templates.models.cisco_aaa_model import CiscoAAAModel
from catalystwan.api.templates.models.cisco_banner_model import CiscoBannerModel
from catalystwan.api.templates.models.cisco_bfd_model import CiscoBFDModel
from catalystwan.api.templates.models.cisco_bgp_model import CiscoBGPModel
from catalystwan.api.templates.models.cisco_logging_model import CiscoLoggingModel
from catalystwan.api.templates.models.cisco_ntp_model import CiscoNTPModel
from catalystwan.api.templates.models.cisco_omp_model import CiscoOMPModel
from catalystwan.api.templates.models.cisco_ospf import CiscoOSPFModel
from catalystwan.api.templates.models.cisco_ospfv3 import CiscoOspfv3Model
from catalystwan.api.templates.models.cisco_secure_internet_gateway import CiscoSecureInternetGatewayModel
from catalystwan.api.templates.models.cisco_snmp_model import CiscoSNMPModel
from catalystwan.api.templates.models.cisco_system import CiscoSystemModel
from catalystwan.api.templates.models.cisco_vpn_interface_model import CiscoVpnInterfaceModel
from catalystwan.api.templates.models.cisco_vpn_model import CiscoVPNModel
from catalystwan.api.templates.models.cli_template import CliTemplateModel
from catalystwan.api.templates.models.omp_vsmart_model import OMPvSmart
from catalystwan.api.templates.models.security_vsmart_model import SecurityvSmart
from catalystwan.api.templates.models.system_vsmart_model import SystemVsmart
from catalystwan.dataclasses import Device, DeviceTemplateInfo, FeatureTemplateInfo, FeatureTemplatesTypes, TemplateInfo
from catalystwan.endpoints.configuration_device_template import FeatureToCLIPayload
from catalystwan.exceptions import AttachedError, TemplateNotFoundError
from catalystwan.response import ManagerResponse
from catalystwan.typed_list import DataSequence
from catalystwan.utils.device_model import DeviceModel
from catalystwan.utils.dict import merge
from catalystwan.utils.pydantic_field import get_extra_field
from catalystwan.utils.template_type import TemplateType

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class DeviceModelError(Exception):
    """Used when unsupported device model used in template."""

    def __init__(self, template, device_models):
        self.message = f"Provided template type '{template.type}' not available for device models: {device_models}"
        super().__init__(self.message)


class DeviceTemplateFeature(Enum):
    LAWFUL_INTERCEPTION = "lawful-interception"
    CLOUD_DOCK = "cloud-dock"
    NETWORK_DESIGN = "network-design"
    VMANAGE_DEFAULT = "vmanage-default"
    ALL = "all"


class TemplatesAPI:
    def __init__(self, session: ManagerSession) -> None:
        self.session = session

    @overload
    def get(self, template: Type[DeviceTemplate]) -> DataSequence[DeviceTemplateInfo]:  # type: ignore
        ...

    @overload
    def get(self, template: Type[FeatureTemplate]) -> DataSequence[FeatureTemplateInfo]:  # type: ignore
        ...

    @overload
    def get(self, template: Type[CLITemplate]) -> DataSequence[TemplateInfo]:  # type: ignore
        ...

    def get(self, template):
        if isinstance(template, FeatureTemplate):
            return self._get_feature_templates()
        if template is FeatureTemplate:
            return self._get_feature_templates()
        if isinstance(template, (DeviceTemplate, CLITemplate)):
            return self._get_device_templates()
        if template in [DeviceTemplate, CLITemplate]:
            return self._get_device_templates()

        raise NotImplementedError()

    def _get_feature_templates(
        self,
        summary: bool = True,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> DataSequence[FeatureTemplateInfo]:
        """In a multitenant vManage system, this API is only available in the Provider view."""
        endpoint = "/dataservice/template/feature"
        params = {"summary": summary}

        fr_templates = self.session.get(url=endpoint, params=params)

        return fr_templates.dataseq(FeatureTemplateInfo)

    def _get_device_templates(
        self, feature: DeviceTemplateFeature = DeviceTemplateFeature.ALL
    ) -> DataSequence[DeviceTemplateInfo]:
        """In a multitenant vManage system, this API is only available in the Provider view."""
        endpoint = "/dataservice/template/device"
        params = {"feature": feature.value}

        templates = self.session.get(url=endpoint, params=params)
        return templates.dataseq(DeviceTemplateInfo)

    def attach(self, name: str, device: Device, timeout_seconds: int = 300, **kwargs):
        template_type = self.get(DeviceTemplate).filter(name=name).single_or_default().config_type
        if template_type == TemplateType.CLI:
            return self._attach_cli(name, device, timeout_seconds=timeout_seconds, **kwargs)

        if template_type == TemplateType.FEATURE:
            return self._attach_feature(name, device, timeout_seconds=timeout_seconds, **kwargs)

        raise NotImplementedError()

    def _attach_feature(self, name: str, device: Device, timeout_seconds: int = 300, **kwargs):
        """Attach Device Template created with Feature Templates.

        Args:
            name: Name of the Device Template to be attached.
            device: Device object under which the template should be attached.
            **device_specific_vars: For parameters in a feature template that you configure as device-specific,
                when you attach a device template to a device, Cisco vManage prompts you for the values to use
                for these parameters. Entering device-specific values in this manner is useful in test or POC networks,
                or if you are deploying a small network. This method generally does not scale well for larger networks.
        """

        def get_device_specific_variables(name: str):
            endpoint = "/dataservice/template/device/config/exportcsv"
            template_id = self.get(DeviceTemplate).filter(name=name).single_or_default().id
            body = {
                "templateId": template_id,
                "isEdited": False,
                "isMasterEdited": False,
            }

            values = self.session.post(endpoint, json=body).json()["header"]["columns"]
            return [DeviceSpecificValue(**value) for value in values]

        vars = get_device_specific_variables(name)
        template_id = self.get(DeviceTemplate).filter(name=name).single_or_default().id
        payload = {
            "deviceTemplateList": [
                {
                    "templateId": template_id,
                    "device": [
                        {
                            "csv-status": "complete",
                            "csv-deviceId": device.uuid,
                            "csv-deviceIP": device.id,
                            "csv-host-name": device.hostname,
                            "csv-templateId": template_id,
                        }
                    ],
                }
            ]
        }

        invalid = False
        for var in vars:
            if var.property not in payload["deviceTemplateList"][0]["device"][0]:
                pointer = payload["deviceTemplateList"][0]["device"][0]
                if var.property not in kwargs["device_specific_vars"]:
                    invalid = True
                    logger.error(f"{var.property} should be provided in attach method as device_specific_vars kwarg.")
                else:
                    pointer[var.property] = kwargs["device_specific_vars"][var.property]  # type: ignore

        if invalid:
            raise TypeError()

        endpoint = "/dataservice/template/device/config/attachfeature"
        logger.info(f"Attaching a template: {name} to the device: {device.hostname}.")
        response = self.session.post(url=endpoint, json=payload).json()
        task = Task(session=self.session, task_id=response["id"]).wait_for_completed(timeout_seconds=timeout_seconds)
        if task.result:
            return True
        logger.warning(f"Failed to attach tempate: {name} to the device: {device.hostname}.")
        logger.warning(f"Task activity information: {task.sub_tasks_data[0].activity}")
        return False

    def _attach_cli(self, name: str, device: Device, is_edited: bool = False, timeout_seconds: int = 300) -> bool:
        """

        Args:
            name (str): Template name to attached.
            device (Device): Device to attach template.
            is_edited (bool): Flag to indicate whether template is being attached as part of edit

        Returns:
            bool: True if attaching template is successful, otherwise - False.
        """
        try:
            template_id = self.get(CLITemplate).filter(name=name).single_or_default().id
            self.template_validation(template_id, device=device)
        except TemplateNotFoundError:
            logger.error(f"Error, Template with name {name} not found on {device}.")
            return False
        payload = {
            "deviceTemplateList": [
                {
                    "templateId": template_id,
                    "device": [
                        {
                            "csv-status": "complete",
                            "csv-deviceId": device.uuid,
                            "csv-deviceIP": device.id,
                            "csv-host-name": device.hostname,
                            "csv-templateId": template_id,
                        }
                    ],
                }
            ]
        }
        if is_edited:
            payload["deviceTemplateList"][0]["isEdited"] = True  # type: ignore
        endpoint = "/dataservice/template/device/config/attachcli"
        logger.info(f"Attaching a template: {name} to the device: {device.hostname}.")
        response = self.session.post(url=endpoint, json=payload).json()
        task = Task(session=self.session, task_id=response["id"]).wait_for_completed(timeout_seconds=timeout_seconds)
        if task.result:
            return True
        logger.warning(f"Failed to attach tempate: {name} to the device: {device.hostname}.")
        logger.warning(f"Task activity information: {task.sub_tasks_data[0].activity}")
        return False

    def deatach(self, device: Device) -> bool:
        """
        Deatach it`s the same to change device mode to CLI mode.

        Args:
            device (Device): Device to deatach template (change mode).

        Returns:
            bool: True if change deatach template (mode to CLI) is successful, otherwise - False.
        """
        payload = {
            "deviceType": device.personality.value,
            "devices": [{"deviceId": device.uuid, "deviceIP": device.id}],
        }
        endpoint = "/dataservice/template/config/device/mode/cli"
        logger.info(f"Changing mode to cli mode for {device.hostname}.")
        response = self.session.post(url=endpoint, json=payload).json()
        task = Task(session=self.session, task_id=response["id"]).wait_for_completed()
        if task.result:
            return True
        logger.warning(f"Failed to change to cli mode for device: {device.hostname}.")
        logger.warning(f"Task activity information: {task.sub_tasks_data[0].activity}")
        return False

    @overload
    def delete(self, template: Type[DeviceTemplate], name: str) -> bool:  # type: ignore
        ...

    @overload
    def delete(self, template: Type[FeatureTemplate], name: str) -> bool:  # type: ignore
        ...

    @overload
    def delete(self, template: Type[CLITemplate], name: str) -> bool:  # type: ignore
        ...

    def delete(self, template, name):
        status = False

        if template is FeatureTemplate:
            status = self._delete_feature_template(name)

        if template is DeviceTemplate and name:
            status = self._delete_device_template(name)

        if isinstance(template, CLITemplate):
            status = self._delete_cli_template(name)

        if template is CLITemplate and name:
            status = self._delete_cli_template(name)

        if status:
            logger.info(f"Template {name} was successfuly deleted.")
            return status

        raise NotImplementedError(f"Not implemented for {template}")

    def _delete_feature_template(self, name: str) -> bool:
        template = self.get(FeatureTemplate).filter(name=name).single_or_default()  # type: ignore
        if template:
            endpoint = f"/dataservice/template/feature/{template.id}"
            self.session.delete(url=endpoint)
        return True

    def _delete_device_template(self, name: str) -> bool:
        """

        Args:
            name (str): Name template to delete.

        Raises:
            AttachedError: If template is attached to device.

        Returns:
            bool: True if deletion is successful, otherwise - False.
        """
        template = self.get(DeviceTemplate).filter(name=name).single_or_default()  # type: ignore
        if template:
            endpoint = f"/dataservice/template/device/{template.id}"
            if template.devices_attached == 0:
                response = self.session.delete(url=endpoint)
                logger.info(f"Template with name: {name} - deleted.")
                return response.ok
            logger.warning(f"Template: {template} is attached to device - cannot be deleted.")
            raise AttachedError(template.name)
        return True

    def _delete_cli_template(self, name: str) -> bool:
        """

        Args:
            name (str): Name template to delete.

        Raises:
            AttachedError: If template is attached to device.

        Returns:
            bool: True if deletion is successful, otherwise - False.
        """
        template = self.get(CLITemplate).filter(name=name).single_or_default()  # type: ignore
        if template:
            endpoint = f"/dataservice/template/device/{template.id}"
            if template.devices_attached == 0:
                response = self.session.delete(url=endpoint)
                logger.info(f"Template with name: {name} - deleted.")
                return response.ok
            logger.warning(f"Template: {template} is attached to device - cannot be deleted.")
            raise AttachedError(template.name)
        return True

    @overload
    def edit(self, template: FeatureTemplate) -> Any:
        ...

    @overload
    def edit(self, template: CLITemplate) -> Any:
        ...

    @overload
    def edit(self, template: DeviceTemplate) -> Any:
        ...

    def edit(self, template):
        template_info = self.get(template).filter(name=template.template_name).single_or_default()
        if not template_info:
            raise TemplateNotFoundError(f"Template with name [{template.template_name}] does not exists.")

        if isinstance(template, FeatureTemplate):
            return self._edit_feature_template(template, template_info)

        if isinstance(template, DeviceTemplate):
            return self._edit_device_template(template)

        raise NotImplementedError()

    def _edit_device_template(self, template: DeviceTemplate):
        self._create_device_template(template, True)

    def _edit_feature_template(self, template: FeatureTemplate, data: FeatureTemplateInfo) -> ManagerResponse:
        if self.is_created_by_generator(template):
            debug = False
            schema = self.get_feature_template_schema(template, debug)
            payload = self.generate_feature_template_payload(template, schema, debug).model_dump(
                by_alias=True, mode="json"
            )
        else:
            payload = json.loads(template.generate_payload(self.session))

        response = self.session.put(f"/dataservice/template/feature/{data.id}", json=payload)
        return response

    @overload
    def create(self, template: FeatureTemplate, debug=False) -> str:
        ...

    @overload
    def create(self, template: DeviceTemplate) -> str:
        ...

    @overload
    def create(self, template: CLITemplate) -> str:
        ...

    def create(self, template, debug: bool = False):
        if isinstance(template, list):
            return [self.create(t) for t in template]

        template_id: Optional[str] = None  # type: ignore
        template_type = None

        # exists = self.get(template).filter(name=template.name)
        # if exists:
        #     raise AlreadyExistsError(f"Template with name [{template.name}] already exists.")

        if isinstance(template, FeatureTemplate):
            if self.is_created_by_generator(template):
                template_id = self.create_by_generator(template, debug)
            else:
                template_id = self._create_feature_template(template)
            template_type = FeatureTemplate.__name__

        if isinstance(template, DeviceTemplate):
            template_id = self._create_device_template(template)
            template_type = DeviceTemplate.__name__

        if isinstance(template, CLITemplate):
            template_id = self._create_cli_template(template)
            template_type = CLITemplate.__name__

        if not template_id:
            raise NotImplementedError()

        logger.info(f"Template {template.template_name} ({template_type}) was created successfully ({template_id}).")
        return template_id

    def _create_feature_template(self, template: FeatureTemplate) -> str:
        payload = template.generate_payload(self.session)
        response = self.session.post("/dataservice/template/feature", json=json.loads(payload))
        template_id = response.json()["templateId"]

        return template_id

    def _create_cli_template(self, template: CLITemplate) -> str:
        payload = template.generate_payload()
        response = self.session.post("/dataservice/template/device/cli/", json=payload)
        template_id = response.json()["templateId"]

        return template_id

    def _create_device_template(self, device_template: DeviceTemplate, edit: bool = False) -> str:
        def get_general_template_info(
            name: str, fr_templates: DataSequence[FeatureTemplateInfo]
        ) -> FeatureTemplateInfo:
            _template = fr_templates.filter(name=name).single_or_default()

            if not _template:
                raise TypeError(f"{name} does not exists. Device Template is invalid.")

            return _template

        def parse_general_template(
            general_template: GeneralTemplate,
            fr_templates: DataSequence[FeatureTemplateInfo],
        ) -> GeneralTemplate:
            if general_template.subTemplates:
                general_template.subTemplates = [
                    parse_general_template(_t, fr_templates) for _t in general_template.subTemplates
                ]
            if general_template.name:
                info = get_general_template_info(general_template.name, fr_templates)
                return GeneralTemplate(
                    name=general_template.name,
                    subTemplates=general_template.subTemplates,
                    templateId=info.id,
                    templateType=info.template_type,
                )
            else:
                return general_template

        fr_templates = self.get(FeatureTemplate)  # type: ignore
        device_template.general_templates = list(
            map(lambda x: parse_general_template(x, fr_templates), device_template.general_templates)  # type: ignore
        )

        if edit:
            template_id = (
                self.session.api.templates.get(DeviceTemplate)
                .filter(name=device_template.template_name)
                .single_or_default()
                .id
            )
            payload = json.loads(device_template.generate_payload())
            response = self.session.put(f"/dataservice/template/device/{template_id}", json=payload)
        else:
            endpoint = "/dataservice/template/device/feature/"
            payload = json.loads(device_template.generate_payload())
            response = self.session.post(endpoint, json=payload)

        return response.text

    def is_created_by_generator(self, template: FeatureTemplate) -> bool:
        """Checks if template is created by generator

        Method will be deleted if every template's payload will be generated dynamically.
        """
        ported_templates = (
            CiscoAAAModel,
            CiscoBFDModel,
            CiscoBannerModel,
            CiscoNTPModel,
            CiscoLoggingModel,
            CiscoOMPModel,
            OMPvSmart,
            SecurityvSmart,
            SystemVsmart,
            CiscoVpnInterfaceModel,
            CiscoSystemModel,
            CiscoSNMPModel,
            CiscoVPNModel,
            CiscoBGPModel,
            CiscoOSPFModel,
            CliTemplateModel,
            CiscoSecureInternetGatewayModel,
            CiscoOspfv3Model,
        )

        return isinstance(template, ported_templates)

    def get_feature_template_schema(self, template: FeatureTemplate, debug: bool = False) -> Any:
        endpoint = f"/dataservice/template/feature/types/definition/{template.type}/15.0.0"
        schema = self.session.get(url=endpoint).json()

        if debug:
            with open(f"response_{template.type}.json", "w") as f:
                f.write(json.dumps(schema, indent=4))

        return schema

    def create_by_generator(self, template: FeatureTemplate, debug: bool) -> str:
        schema = self.get_feature_template_schema(template, debug)
        payload = self.generate_feature_template_payload(template, schema, debug)

        endpoint = "/dataservice/template/feature"
        response = self.session.post(endpoint, json=payload.model_dump(by_alias=True, exclude_none=True, mode="json"))

        return response.json()["templateId"]

    def generate_feature_template_payload(
        self, template: FeatureTemplate, schema: Any, debug: bool = False
    ) -> FeatureTemplatePayload:
        payload = FeatureTemplatePayload(
            name=template.template_name,
            description=template.template_description,
            template_type=template.type,
            device_types=[device_model.value for device_model in template.device_models],
            definition={},
        )  # type: ignore

        fr_template_fields = [FeatureTemplateField(**field) for field in schema["fields"]]  # TODO
        json_dumped_template = template.model_dump(mode="json")
        # "name"
        for field in fr_template_fields:
            value = None
            priority_order = None
            # TODO How to discover Device specific variable
            if field.key in template.device_specific_variables:
                value = template.device_specific_variables[field.key]
            else:
                for field_name, field_value in template.model_fields.items():
                    data_path = get_extra_field(field_value, "data_path", default=[])
                    vmanage_key = get_extra_field(field_value, "vmanage_key")
                    if field.dataPath == data_path and (  # type: ignore
                        (field.key == field_value.alias or field.key == field_name)
                        or field.key == vmanage_key  # type: ignore
                    ):
                        priority_order = get_extra_field(field_value, "priority_order")  # type: ignore
                        value = getattr(template, field_name)
                        json_dumped_value = json_dumped_template.get(field_name)
                        break
                if value is None:
                    continue

            payload.definition = merge(
                payload.definition,
                field.payload_scheme(value, json_dumped_value=json_dumped_value, priority_order=priority_order),
            )

        if debug:
            with open(f"payload_{template.type}.json", "w") as f:
                f.write(json.dumps(payload.model_dump(by_alias=True, mode="json"), indent=4))

        return payload

    def validate_device_model(self, template: FeatureTemplate) -> bool:
        """Verify if selected template can be used with provided device model"""

        template_type = self._get_feature_template_types().filter(name=template.type).single_or_default()

        available_devices_for_template = [device["name"] for device in template_type.device_models]

        provided_device_models = [
            dev_mod.value if type(dev_mod) is DeviceModel else dev_mod for dev_mod in template.device_models
        ]

        if not all(dev in available_devices_for_template for dev in provided_device_models):
            logger.debug(f"Available devices for template '{template.type}': {available_devices_for_template}")
            raise DeviceModelError(template, provided_device_models)
        return True

    def _get_feature_template_types(self, type: str = "all") -> DataSequence[FeatureTemplatesTypes]:
        """Gets list off all templates and devices associated with these templates"""

        endpoint = "/dataservice/template/feature/types"
        params = {"type": type}
        response = self.session.get(endpoint, params=params)

        return response.dataseq(FeatureTemplatesTypes)

    def template_validation(self, id: str, device: Device) -> str:
        """Checking the template of the configuration on the machine.

        Args:
            id (str): template id to check.
            device (Device): The device on which the configuration is to be validate.

        Returns:
            str: Validated config.
        """
        payload = {
            "templateId": id,
            "device": {
                "csv-status": "complete",
                "csv-deviceId": device.uuid,
                "csv-deviceIP": device.id,
                "csv-host-name": device.hostname,
                "csv-templateId": id,
            },
            "isEdited": False,
            "isMasterEdited": False,
            "isRFSRequired": True,
        }
        endpoint = "/dataservice/template/device/config/config/"
        response = self.session.post(url=endpoint, json=payload)
        return response.text

    def edit_before_push(self, name: str, device: Device) -> bool:
        """
        Edits device / CLI template before pushing modified config to device(s)

        Args:
            name (str): Template name to edit.
            device (Device): Device to attach template.

        Returns:
            bool: True if edit template is successful, otherwise - False.
        """
        try:
            template_id = self.get(CLITemplate).filter(name=name).single_or_default().id
            self.template_validation(template_id, device=device)
        except TemplateNotFoundError:
            logger.error(f"Error, Template with name {name} not found on {device}.")
            return False
        payload = {
            "templateId": template_id,
            "deviceIds": [device.uuid],
            "isEdited": True,
            "isMasterEdited": True,
        }
        endpoint = "/dataservice/template/device/config/input/"
        logger.info(f"Editing template: {name} of device: {device.hostname}.")
        response = self.session.post(url=endpoint, json=payload).json()
        if (response.get("data") is not None) and (response["data"][0].get("csv-status") == "complete"):
            return True
        logger.warning(f"Failed to edit tempate: {name} of device: {device.hostname}.")
        return False

    def get_device_configuration_preview(self, payload: FeatureToCLIPayload) -> CiscoConfParse:
        text_config = self.session.endpoints.configuration_device_template.get_device_configuration_preview(payload)

        return CiscoConfParse(text_config.splitlines())

    def load_running(self, device: Device) -> CiscoConfParse:
        """Load running config from device.

        Args:
            device: The device from which load config.

        Returns:
            CiscoConfParse: A working configuration on the machine.
        """
        encoded_uuid = device.uuid.replace("/", "%2F")
        endpoint = f"/dataservice/template/config/running/{encoded_uuid}"
        response = self.session.get_json(endpoint)
        config = CiscoConfParse(response["config"].splitlines())
        logger.debug(f"Template loaded from {device.hostname}.")
        return config



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# type: ignore
import json
import unittest
from pathlib import Path
from typing import Any, Dict, List
from unittest import TestCase
from unittest.mock import patch

import pytest
from parameterized import parameterized

import catalystwan.tests.templates.models as models
from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.dataclasses import FeatureTemplateInfo
from catalystwan.typed_list import DataSequence
from catalystwan.utils.creation_tools import create_dataclass


class TestFeatureTemplate(TestCase):
    def setUp(self):
        template: FeatureTemplate
        feature_template_response: List[Dict[str, Any]] = []

        for template in map(models.__dict__.get, models.__all__):
            definition: Dict[str, Any]
            with open(
                Path(__file__).resolve().parents[0] / Path("definitions") / Path(template.template_name + ".json")
            ) as f:
                definition = json.load(f)
            feature_template_response.append(
                {
                    "last_updated_by": "admin",
                    "resource_group": "global",
                    "id": "xxx",
                    "factory_default": "False",
                    "name": template.template_name,
                    "devices_attached": 1,
                    "description": template.template_description,
                    "last_updated_on": 1111111111111,
                    "template_type": template.type,
                    "device_type": ["vedge-C8000V"],
                    "version": "15.0.0",
                    "template_definiton": f"{json.dumps(definition)}",
                }
            )

        self.get_feature_templates_response = DataSequence(
            FeatureTemplateInfo,
            [create_dataclass(FeatureTemplateInfo, response) for response in feature_template_response],
        )

    @parameterized.expand([(template,) for template in map(models.__dict__.get, models.__all__)])
    @pytest.mark.skip(reason="Deserialization to be refactored")
    @patch("catalystwan.session.ManagerSession")
    def test_get(self, template: FeatureTemplate, mock_session):
        # Arrange
        mock_session.api.templates._get_feature_templates.return_value = self.get_feature_templates_response

        # Act
        feature_template_from_get = FeatureTemplate.get(session=mock_session, name=template.template_name)

        # Assert
        self.assertEqual(feature_template_from_get, template)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum


class Timezone(str, Enum):
    EUROPE_ANDORRA = "Europe/Andorra"
    ASIA_DUBAI = "Asia/Dubai"
    ASIA_KABUL_ = "Asia/Kabul"
    AMERICA_ANTIGUA_ = "America/Antigua"
    AMERICA_ANGUILLA_ = "America/Anguilla"
    EUROPE_TIRANE_ = "Europe/Tirane"
    ASIA_YEREVAN_ = "Asia/Yerevan"
    AFRICA_LUANDA_ = "Africa/Luanda"
    ANTARCTICA_MCMURDO = "Antarctica/McMurdo"
    ANTARCTICA_ROTHERA = "Antarctica/Rothera"
    ANTARCTICA_PALMER = "Antarctica/Palmer"
    ANTARCTICA_MAWSON = "Antarctica/Mawson"
    ANTARCTICA_DAVIS = "Antarctica/Davis"
    ANTARCTICA_CASEY = "Antarctica/Casey"
    ANTARCTICA_VOSTOK = "Antarctica/Vostok"
    ANTARCTICA_DUMONTDURVILLE = "Antarctica/DumontDUrville"
    ANTARCTICA_SYOWA = "Antarctica/Syowa"
    AMERICA_ARGENTINA_BUENOS_AIRES = "America/Argentina/Buenos_Aires"
    AMERICA_ARGENTINA_CORDOBA = "America/Argentina/Cordoba"
    AMERICA_ARGENTINA_SALTA = "America/Argentina/Salta"
    AMERICA_ARGENTINA_JUJUY = "America/Argentina/Jujuy"
    AMERICA_ARGENTINA_TUCUMAN = "America/Argentina/Tucuman"
    AMERICA_ARGENTINA_CATAMARCA = "America/Argentina/Catamarca"
    AMERICA_ARGENTINA_LA_RIOJA = "America/Argentina/La_Rioja"
    AMERICA_ARGENTINA_SAN_JUAN = "America/Argentina/San_Juan"
    AMERICA_ARGENTINA_MENDOZA = "America/Argentina/Mendoza"
    AMERICA_ARGENTINA_SAN_LUIS = "America/Argentina/San_Luis"
    AMERICA_ARGENTINA_RIO_GALLEGOS = "America/Argentina/Rio_Gallegos"
    AMERICA_ARGENTINA_USHUAIA = "America/Argentina/Ushuaia"
    PACIFIC_PAGO_PAGO = "Pacific/Pago_Pago"
    EUROPE_VIENNA = "Europe/Vienna"
    AUSTRALIA_LORD_HOWE = "Australia/Lord_Howe"
    ANTARCTICA_MACQUARIE = "Antarctica/Macquarie"
    AUSTRALIA_HOBART = "Australia/Hobart"
    AUSTRALIA_CURRIE = "Australia/Currie"
    AUSTRALIA_MELBOURNE = "Australia/Melbourne"
    AUSTRALIA_SYDNEY = "Australia/Sydney"
    AUSTRALIA_BROKEN_HILL = "Australia/Broken_Hill"
    AUSTRALIA_BRISBANE = "Australia/Brisbane"
    AUSTRALIA_LINDEMAN = "Australia/Lindeman"
    AUSTRALIA_ADELAIDE = "Australia/Adelaide"
    AUSTRALIA_DARWIN = "Australia/Darwin"
    AUSTRALIA_PERTH = "Australia/Perth"
    AUSTRALIA_EUCLA = "Australia/Eucla"
    AMERICA_ARUBA = "America/Aruba"
    EUROPE_MARIEHAMN = "Europe/Mariehamn"
    ASIA_BAKU = "Asia/Baku"
    EUROPE_SARAJEVO = "Europe/Sarajevo"
    AMERICA_BARBADOS = "America/Barbados"
    ASIA_DHAKA = "Asia/Dhaka"
    EUROPE_BRUSSELS = "Europe/Brussels"
    AFRICA_OUAGADOUGOU = "Africa/Ouagadougou"
    EUROPE_SOFIA = "Europe/Sofia"
    ASIA_BAHRAIN = "Asia/Bahrain"
    AFRICA_BUJUMBURA = "Africa/Bujumbura"
    AFRICA_PORTO_NOVO = "Africa/Porto-Novo"
    AMERICA_ST_BARTHELEMY = "America/St_Barthelemy"
    ATLANTIC_BERMUDA = "Atlantic/Bermuda"
    ASIA_BRUNEI = "Asia/Brunei"
    AMERICA_LA_PAZ = "America/La_Paz"
    AMERICA_KRALENDIJK = "America/Kralendijk"
    AMERICA_NORONHA = "America/Noronha"
    AMERICA_BELEM = "America/Belem"
    AMERICA_FORTALEZA = "America/Fortaleza"
    AMERICA_RECIFE = "America/Recife"
    AMERICA_ARAGUAINA = "America/Araguaina"
    AMERICA_MACEIO = "America/Maceio"
    AMERICA_BAHIA = "America/Bahia"
    AMERICA_SAO_PAULO = "America/Sao_Paulo"
    AMERICA_CAMPO_GRANDE = "America/Campo_Grande"
    AMERICA_CUIABA = "America/Cuiaba"
    AMERICA_SANTAREM = "America/Santarem"
    AMERICA_PORTO_VELHO = "America/Porto_Velho"
    AMERICA_BOA_VISTA = "America/Boa_Vista"
    AMERICA_MANAUS = "America/Manaus"
    AMERICA_EIRUNEPE = "America/Eirunepe"
    AMERICA_RIO_BRANCO = "America/Rio_Branco"
    AMERICA_NASSAU = "America/Nassau"
    ASIA_THIMPHU = "Asia/Thimphu"
    AFRICA_GABORONE = "Africa/Gaborone"
    EUROPE_MINSK = "Europe/Minsk"
    AMERICA_BELIZE = "America/Belize"
    AMERICA_ST_JOHNS = "America/St_Johns"
    AMERICA_HALIFAX = "America/Halifax"
    AMERICA_GLACE_BAY = "America/Glace_Bay"
    AMERICA_MONCTON = "America/Moncton"
    AMERICA_GOOSE_BAY = "America/Goose_Bay"
    AMERICA_BLANC_SABLON = "America/Blanc-Sablon"
    AMERICA_TORONTO = "America/Toronto"
    AMERICA_NIPIGON = "America/Nipigon"
    AMERICA_THUNDER_BAY = "America/Thunder_Bay"
    AMERICA_IQALUIT = "America/Iqaluit"
    AMERICA_PANGNIRTUNG = "America/Pangnirtung"
    AMERICA_RESOLUTE = "America/Resolute"
    AMERICA_ATIKOKAN = "America/Atikokan"
    AMERICA_RANKIN_INLET = "America/Rankin_Inlet"
    AMERICA_WINNIPEG = "America/Winnipeg"
    AMERICA_RAINY_RIVER = "America/Rainy_River"
    AMERICA_REGINA = "America/Regina"
    AMERICA_SWIFT_CURRENT = "America/Swift_Current"
    AMERICA_EDMONTON = "America/Edmonton"
    AMERICA_CAMBRIDGE_BAY = "America/Cambridge_Bay"
    AMERICA_YELLOWKNIFE = "America/Yellowknife"
    AMERICA_INUVIK = "America/Inuvik"
    AMERICA_CRESTON = "America/Creston"
    AMERICA_DAWSON_CREEK = "America/Dawson_Creek"
    AMERICA_VANCOUVER = "America/Vancouver"
    AMERICA_WHITEHORSE = "America/Whitehorse"
    AMERICA_DAWSON = "America/Dawson"
    INDIAN_COCOS = "Indian/Cocos"
    AFRICA_KINSHASA = "Africa/Kinshasa"
    AFRICA_LUBUMBASHI = "Africa/Lubumbashi"
    AFRICA_BANGUI = "Africa/Bangui"
    AFRICA_BRAZZAVILLE = "Africa/Brazzaville"
    EUROPE_ZURICH = "Europe/Zurich"
    AFRICA_ABIDJAN = "Africa/Abidjan"
    PACIFIC_RAROTONGA = "Pacific/Rarotonga"
    AMERICA_SANTIAGO = "America/Santiago"
    PACIFIC_EASTER = "Pacific/Easter"
    AFRICA_DOUALA = "Africa/Douala"
    ASIA_SHANGHAI = "Asia/Shanghai"
    ASIA_HARBIN = "Asia/Harbin"
    ASIA_CHONGQING = "Asia/Chongqing"
    ASIA_URUMQI = "Asia/Urumqi"
    ASIA_KASHGAR = "Asia/Kashgar"
    AMERICA_BOGOTA = "America/Bogota"
    AMERICA_COSTA_RICA = "America/Costa_Rica"
    AMERICA_HAVANA = "America/Havana"
    ATLANTIC_CAPE_VERDE = "Atlantic/Cape_Verde"
    AMERICA_CURACAO = "America/Curacao"
    INDIAN_CHRISTMAS = "Indian/Christmas"
    ASIA_NICOSIA = "Asia/Nicosia"
    EUROPE_PRAGUE = "Europe/Prague"
    EUROPE_BERLIN = "Europe/Berlin"
    EUROPE_BUSINGEN = "Europe/Busingen"
    AFRICA_DJIBOUTI = "Africa/Djibouti"
    EUROPE_COPENHAGEN = "Europe/Copenhagen"
    AMERICA_DOMINICA = "America/Dominica"
    AMERICA_SANTO_DOMINGO = "America/Santo_Domingo"
    AFRICA_ALGIERS = "Africa/Algiers"
    AMERICA_GUAYAQUIL = "America/Guayaquil"
    PACIFIC_GALAPAGOS = "Pacific/Galapagos"
    EUROPE_TALLINN = "Europe/Tallinn"
    AFRICA_CAIRO = "Africa/Cairo"
    AFRICA_EL_AAIUN = "Africa/El_Aaiun"
    AFRICA_ASMARA = "Africa/Asmara"
    EUROPE_MADRID = "Europe/Madrid"
    AFRICA_CEUTA = "Africa/Ceuta"
    ATLANTIC_CANARY = "Atlantic/Canary"
    AFRICA_ADDIS_ABABA = "Africa/Addis_Ababa"
    EUROPE_HELSINKI = "Europe/Helsinki"
    PACIFIC_FIJI = "Pacific/Fiji"
    ATLANTIC_STANLEY = "Atlantic/Stanley"
    PACIFIC_CHUUK = "Pacific/Chuuk"
    PACIFIC_POHNPEI = "Pacific/Pohnpei"
    PACIFIC_KOSRAE = "Pacific/Kosrae"
    ATLANTIC_FAROE = "Atlantic/Faroe"
    EUROPE_PARIS = "Europe/Paris"
    AFRICA_LIBREVILLE = "Africa/Libreville"
    EUROPE_LONDON = "Europe/London"
    AMERICA_GRENADA = "America/Grenada"
    ASIA_TBILISI = "Asia/Tbilisi"
    AMERICA_CAYENNE = "America/Cayenne"
    EUROPE_GUERNSEY = "Europe/Guernsey"
    AFRICA_ACCRA = "Africa/Accra"
    EUROPE_GIBRALTAR = "Europe/Gibraltar"
    AMERICA_GODTHAB = "America/Godthab"
    AMERICA_DANMARKSHAVN = "America/Danmarkshavn"
    AMERICA_SCORESBYSUND = "America/Scoresbysund"
    AMERICA_THULE = "America/Thule"
    AFRICA_BANJUL = "Africa/Banjul"
    AFRICA_CONAKRY = "Africa/Conakry"
    AMERICA_GUADELOUPE = "America/Guadeloupe"
    AFRICA_MALABO = "Africa/Malabo"
    EUROPE_ATHENS = "Europe/Athens"
    ATLANTIC_SOUTH_GEORGIA = "Atlantic/South_Georgia"
    AMERICA_GUATEMALA = "America/Guatemala"
    PACIFIC_GUAM = "Pacific/Guam"
    AFRICA_BISSAU = "Africa/Bissau"
    AMERICA_GUYANA = "America/Guyana"
    ASIA_HONG_KONG = "Asia/Hong_Kong"
    AMERICA_TEGUCIGALPA = "America/Tegucigalpa"
    EUROPE_ZAGREB = "Europe/Zagreb"
    AMERICA_PORT_AU_PRINCE = "America/Port-au-Prince"
    EUROPE_BUDAPEST = "Europe/Budapest"
    ASIA_JAKARTA = "Asia/Jakarta"
    ASIA_PONTIANAK = "Asia/Pontianak"
    ASIA_MAKASSAR = "Asia/Makassar"
    ASIA_JAYAPURA = "Asia/Jayapura"
    EUROPE_DUBLIN = "Europe/Dublin"
    ASIA_JERUSALEM = "Asia/Jerusalem"
    EUROPE_ISLE_OF_MAN = "Europe/Isle_of_Man"
    ASIA_KOLKATA = "Asia/Kolkata"
    INDIAN_CHAGOS = "Indian/Chagos"
    ASIA_BAGHDAD = "Asia/Baghdad"
    ASIA_TEHRAN = "Asia/Tehran"
    ATLANTIC_REYKJAVIK = "Atlantic/Reykjavik"
    EUROPE_ROME = "Europe/Rome"
    EUROPE_JERSEY = "Europe/Jersey"
    AMERICA_JAMAICA = "America/Jamaica"
    ASIA_AMMAN = "Asia/Amman"
    ASIA_TOKYO = "Asia/Tokyo"
    AFRICA_NAIROBI = "Africa/Nairobi"
    ASIA_BISHKEK = "Asia/Bishkek"
    ASIA_PHNOM_PENH = "Asia/Phnom_Penh"
    PACIFIC_TARAWA = "Pacific/Tarawa"
    PACIFIC_ENDERBURY = "Pacific/Enderbury"
    PACIFIC_KIRITIMATI = "Pacific/Kiritimati"
    INDIAN_COMORO = "Indian/Comoro"
    AMERICA_ST_KITTS = "America/St_Kitts"
    ASIA_PYONGYANG = "Asia/Pyongyang"
    ASIA_SEOUL = "Asia/Seoul"
    ASIA_KUWAIT = "Asia/Kuwait"
    AMERICA_CAYMAN = "America/Cayman"
    ASIA_ALMATY = "Asia/Almaty"
    ASIA_QYZYLORDA = "Asia/Qyzylorda"
    ASIA_AQTOBE = "Asia/Aqtobe"
    ASIA_AQTAU = "Asia/Aqtau"
    ASIA_ORAL = "Asia/Oral"
    ASIA_VIENTIANE = "Asia/Vientiane"
    ASIA_BEIRUT = "Asia/Beirut"
    AMERICA_ST_LUCIA = "America/St_Lucia"
    EUROPE_VADUZ = "Europe/Vaduz"
    ASIA_COLOMBO = "Asia/Colombo"
    AFRICA_MONROVIA = "Africa/Monrovia"
    AFRICA_MASERU = "Africa/Maseru"
    EUROPE_VILNIUS = "Europe/Vilnius"
    EUROPE_LUXEMBOURG = "Europe/Luxembourg"
    EUROPE_RIGA = "Europe/Riga"
    AFRICA_TRIPOLI = "Africa/Tripoli"
    AFRICA_CASABLANCA = "Africa/Casablanca"
    EUROPE_MONACO = "Europe/Monaco"
    EUROPE_CHISINAU = "Europe/Chisinau"
    EUROPE_PODGORICA = "Europe/Podgorica"
    AMERICA_MARIGOT = "America/Marigot"
    INDIAN_ANTANANARIVO = "Indian/Antananarivo"
    PACIFIC_MAJURO = "Pacific/Majuro"
    PACIFIC_KWAJALEIN = "Pacific/Kwajalein"
    EUROPE_SKOPJE = "Europe/Skopje"
    AFRICA_BAMAKO = "Africa/Bamako"
    ASIA_RANGOON = "Asia/Rangoon"
    ASIA_ULAANBAATAR = "Asia/Ulaanbaatar"
    ASIA_HOVD = "Asia/Hovd"
    ASIA_CHOIBALSAN = "Asia/Choibalsan"
    ASIA_MACAU = "Asia/Macau"
    PACIFIC_SAIPAN = "Pacific/Saipan"
    AMERICA_MARTINIQUE = "America/Martinique"
    AFRICA_NOUAKCHOTT = "Africa/Nouakchott"
    AMERICA_MONTSERRAT = "America/Montserrat"
    EUROPE_MALTA = "Europe/Malta"
    INDIAN_MAURITIUS = "Indian/Mauritius"
    INDIAN_MALDIVES = "Indian/Maldives"
    AFRICA_BLANTYRE = "Africa/Blantyre"
    AMERICA_MEXICO_CITY = "America/Mexico_City"
    AMERICA_CANCUN = "America/Cancun"
    AMERICA_MERIDA = "America/Merida"
    AMERICA_MONTERREY = "America/Monterrey"
    AMERICA_MATAMOROS = "America/Matamoros"
    AMERICA_MAZATLAN = "America/Mazatlan"
    AMERICA_CHIHUAHUA = "America/Chihuahua"
    AMERICA_OJINAGA = "America/Ojinaga"
    AMERICA_HERMOSILLO = "America/Hermosillo"
    AMERICA_TIJUANA = "America/Tijuana"
    AMERICA_SANTA_ISABEL = "America/Santa_Isabel"
    AMERICA_BAHIA_BANDERAS = "America/Bahia_Banderas"
    ASIA_KUALA_LUMPUR = "Asia/Kuala_Lumpur"
    ASIA_KUCHING = "Asia/Kuching"
    AFRICA_MAPUTO = "Africa/Maputo"
    AFRICA_WINDHOEK = "Africa/Windhoek"
    PACIFIC_NOUMEA = "Pacific/Noumea"
    AFRICA_NIAMEY = "Africa/Niamey"
    PACIFIC_NORFOLK = "Pacific/Norfolk"
    AFRICA_LAGOS = "Africa/Lagos"
    AMERICA_MANAGUA = "America/Managua"
    EUROPE_AMSTERDAM = "Europe/Amsterdam"
    EUROPE_OSLO = "Europe/Oslo"
    ASIA_KATHMANDU = "Asia/Kathmandu"
    PACIFIC_NAURU = "Pacific/Nauru"
    PACIFIC_NIUE = "Pacific/Niue"
    PACIFIC_AUCKLAND = "Pacific/Auckland"
    PACIFIC_CHATHAM = "Pacific/Chatham"
    ASIA_MUSCAT = "Asia/Muscat"
    AMERICA_PANAMA = "America/Panama"
    AMERICA_LIMA = "America/Lima"
    PACIFIC_TAHITI = "Pacific/Tahiti"
    PACIFIC_MARQUESAS = "Pacific/Marquesas"
    PACIFIC_GAMBIER = "Pacific/Gambier"
    PACIFIC_PORT_MORESBY = "Pacific/Port_Moresby"
    ASIA_MANILA = "Asia/Manila"
    ASIA_KARACHI = "Asia/Karachi"
    EUROPE_WARSAW = "Europe/Warsaw"
    AMERICA_MIQUELON = "America/Miquelon"
    PACIFIC_PITCAIRN = "Pacific/Pitcairn"
    AMERICA_PUERTO_RICO = "America/Puerto_Rico"
    ASIA_GAZA = "Asia/Gaza"
    ASIA_HEBRON = "Asia/Hebron"
    EUROPE_LISBON = "Europe/Lisbon"
    ATLANTIC_MADEIRA = "Atlantic/Madeira"
    ATLANTIC_AZORES = "Atlantic/Azores"
    PACIFIC_PALAU = "Pacific/Palau"
    AMERICA_ASUNCION = "America/Asuncion"
    ASIA_QATAR = "Asia/Qatar"
    INDIAN_REUNION = "Indian/Reunion"
    EUROPE_BUCHAREST = "Europe/Bucharest"
    EUROPE_BELGRADE = "Europe/Belgrade"
    EUROPE_KALININGRAD = "Europe/Kaliningrad"
    EUROPE_MOSCOW = "Europe/Moscow"
    EUROPE_VOLGOGRAD = "Europe/Volgograd"
    EUROPE_SAMARA = "Europe/Samara"
    ASIA_YEKATERINBURG = "Asia/Yekaterinburg"
    ASIA_OMSK = "Asia/Omsk"
    ASIA_NOVOSIBIRSK = "Asia/Novosibirsk"
    ASIA_NOVOKUZNETSK = "Asia/Novokuznetsk"
    ASIA_KRASNOYARSK = "Asia/Krasnoyarsk"
    ASIA_IRKUTSK = "Asia/Irkutsk"
    ASIA_YAKUTSK = "Asia/Yakutsk"
    ASIA_KHANDYGA = "Asia/Khandyga"
    ASIA_VLADIVOSTOK = "Asia/Vladivostok"
    ASIA_SAKHALIN = "Asia/Sakhalin"
    ASIA_UST_NERA = "Asia/Ust-Nera"
    ASIA_MAGADAN = "Asia/Magadan"
    ASIA_KAMCHATKA = "Asia/Kamchatka"
    ASIA_ANADYR = "Asia/Anadyr"
    AFRICA_KIGALI = "Africa/Kigali"
    ASIA_RIYADH = "Asia/Riyadh"
    PACIFIC_GUADALCANAL = "Pacific/Guadalcanal"
    INDIAN_MAHE = "Indian/Mahe"
    AFRICA_KHARTOUM = "Africa/Khartoum"
    EUROPE_STOCKHOLM = "Europe/Stockholm"
    ASIA_SINGAPORE = "Asia/Singapore"
    ATLANTIC_ST_HELENA = "Atlantic/St_Helena"
    EUROPE_LJUBLJANA = "Europe/Ljubljana"
    ARCTIC_LONGYEARBYEN = "Arctic/Longyearbyen"
    EUROPE_BRATISLAVA = "Europe/Bratislava"
    AFRICA_FREETOWN = "Africa/Freetown"
    EUROPE_SAN_MARINO = "Europe/San_Marino"
    AFRICA_DAKAR = "Africa/Dakar"
    AFRICA_MOGADISHU = "Africa/Mogadishu"
    AMERICA_PARAMARIBO = "America/Paramaribo"
    AFRICA_JUBA = "Africa/Juba"
    AFRICA_SAO_TOME = "Africa/Sao_Tome"
    AMERICA_EL_SALVADOR = "America/El_Salvador"
    AMERICA_LOWER_PRINCES = "America/Lower_Princes"
    ASIA_DAMASCUS = "Asia/Damascus"
    AFRICA_MBABANE = "Africa/Mbabane"
    AMERICA_GRAND_TURK = "America/Grand_Turk"
    AFRICA_NDJAMENA = "Africa/Ndjamena"
    INDIAN_KERGUELEN = "Indian/Kerguelen"
    AFRICA_LOME = "Africa/Lome"
    ASIA_BANGKOK = "Asia/Bangkok"
    ASIA_DUSHANBE = "Asia/Dushanbe"
    PACIFIC_FAKAOFO = "Pacific/Fakaofo"
    ASIA_DILI = "Asia/Dili"
    ASIA_ASHGABAT = "Asia/Ashgabat"
    AFRICA_TUNIS = "Africa/Tunis"
    PACIFIC_TONGATAPU = "Pacific/Tongatapu"
    EUROPE_ISTANBUL = "Europe/Istanbul"
    AMERICA_PORT_OF_SPAIN = "America/Port_of_Spain"
    PACIFIC_FUNAFUTI = "Pacific/Funafuti"
    ASIA_TAIPEI = "Asia/Taipei"
    AFRICA_DAR_ES_SALAAM = "Africa/Dar_es_Salaam"
    EUROPE_KIEV = "Europe/Kiev"
    EUROPE_UZHGOROD = "Europe/Uzhgorod"
    EUROPE_ZAPOROZHYE = "Europe/Zaporozhye"
    EUROPE_SIMFEROPOL = "Europe/Simferopol"
    AFRICA_KAMPALA = "Africa/Kampala"
    PACIFIC_JOHNSTON = "Pacific/Johnston"
    PACIFIC_MIDWAY = "Pacific/Midway"
    PACIFIC_WAKE = "Pacific/Wake"
    AMERICA_NEW_YORK = "America/New_York"
    AMERICA_DETROIT = "America/Detroit"
    AMERICA_KENTUCKY_LOUISVILLE = "America/Kentucky/Louisville"
    AMERICA_KENTUCKY_MONTICELLO = "America/Kentucky/Monticello"
    AMERICA_INDIANA_INDIANAPOLIS = "America/Indiana/Indianapolis"
    AMERICA_INDIANA_VINCENNES = "America/Indiana/Vincennes"
    AMERICA_INDIANA_WINAMAC = "America/Indiana/Winamac"
    AMERICA_INDIANA_MARENGO = "America/Indiana/Marengo"
    AMERICA_INDIANA_PETERSBURG = "America/Indiana/Petersburg"
    AMERICA_INDIANA_VEVAY = "America/Indiana/Vevay"
    AMERICA_CHICAGO = "America/Chicago"
    AMERICA_INDIANA_TELL_CITY = "America/Indiana/Tell_City"
    AMERICA_INDIANA_KNOX = "America/Indiana/Knox"
    AMERICA_MENOMINEE = "America/Menominee"
    AMERICA_NORTH_DAKOTA_CENTER = "America/North_Dakota/Center"
    AMERICA_NORTH_DAKOTA_NEW_SALEM = "America/North_Dakota/New_Salem"
    AMERICA_NORTH_DAKOTA_BEULAH = "America/North_Dakota/Beulah"
    AMERICA_DENVER = "America/Denver"
    AMERICA_BOISE = "America/Boise"
    AMERICA_PHOENIX = "America/Phoenix"
    AMERICA_LOS_ANGELES = "America/Los_Angeles"
    AMERICA_ANCHORAGE = "America/Anchorage"
    AMERICA_JUNEAU = "America/Juneau"
    AMERICA_SITKA = "America/Sitka"
    AMERICA_YAKUTAT = "America/Yakutat"
    AMERICA_NOME = "America/Nome"
    AMERICA_ADAK = "America/Adak"
    AMERICA_METLAKATLA = "America/Metlakatla"
    PACIFIC_HONOLULU = "Pacific/Honolulu"
    AMERICA_MONTEVIDEO = "America/Montevideo"
    ASIA_SAMARKAND = "Asia/Samarkand"
    ASIA_TASHKENT = "Asia/Tashkent"
    EUROPE_VATICAN = "Europe/Vatican"
    AMERICA_ST_VINCENT = "America/St_Vincent"
    AMERICA_CARACAS = "America/Caracas"
    AMERICA_TORTOLA = "America/Tortola"
    AMERICA_ST_THOMAS = "America/St_Thomas"
    ASIA_HO_CHI_MINH = "Asia/Ho_Chi_Minh"
    PACIFIC_EFATE = "Pacific/Efate"
    PACIFIC_WALLIS = "Pacific/Wallis"
    PACIFIC_APIA = "Pacific/Apia"
    ASIA_ADEN = "Asia/Aden"
    INDIAN_MAYOTTE = "Indian/Mayotte"
    AFRICA_JOHANNESBURG = "Africa/Johannesburg"
    AFRICA_LUSAKA = "Africa/Lusaka"
    AFRICA_HARARE = "Africa/Harare"
    UTC = "UTC"



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Any, Optional

from pydantic.fields import FieldInfo


def get_extra_field(field_info: FieldInfo, key: str, default: Optional[Any] = None) -> Any:
    try:
        return field_info.json_schema_extra.get(key, default)  # type: ignore
    except AttributeError:
        return default



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, IPvAnyAddress, field_validator

from catalystwan.models.policy.policy import AssemblyItemBase, PolicyCreationPayload, PolicyDefinition, PolicyInfo

LocalizedPolicySupportedItemType = Literal[
    "qosMap",
    "rewriteRule",
    "vpnQoSMap",
    "acl",
    "aclv6",
    "deviceaccesspolicy",
    "deviceAccessPolicy",
    "deviceaccesspolicyv6",
    "deviceAccessPolicyv6",
    "vedgeRoute",
]


class LocalizedPolicySettings(BaseModel):
    flow_visibility: Optional[bool] = Field(
        default=None, serialization_alias="flowVisibility", validation_alias="flowVisibility"
    )
    flow_visibility_ipv6: Optional[bool] = Field(
        default=None, serialization_alias="flowVisibilityIPv6", validation_alias="flowVisibilityIPv6"
    )
    app_visibility: Optional[bool] = Field(
        default=None, serialization_alias="appVisibility", validation_alias="appVisibility"
    )
    app_visibility_ipv6: Optional[bool] = Field(
        default=None, serialization_alias="appVisibilityIPv6", validation_alias="appVisibilityIPv6"
    )
    cloud_qos: Optional[bool] = Field(default=None, serialization_alias="cloudQos", validation_alias="cloudQos")
    cloud_qos_service_side: Optional[bool] = Field(
        default=None, serialization_alias="cloudQosServiceSide", validation_alias="cloudQosServiceSide"
    )
    implicit_acl_logging: Optional[bool] = Field(
        default=None, serialization_alias="implicitAclLogging", validation_alias="implicitAclLogging"
    )
    log_frequency: Optional[int] = Field(
        default=None, serialization_alias="logFrequency", validation_alias="logFrequency", ge=0, le=2147483647
    )
    ip_visibility_cache_entries: Optional[int] = Field(
        default=None,
        serialization_alias="ipVisibilityCacheEntries",
        validation_alias="ipVisibilityCacheEntries",
        ge=16,
        le=2000000,
    )
    ip_v6_visibility_cache_entries: Optional[int] = Field(
        default=None,
        serialization_alias="ipV6VisibilityCacheEntries",
        validation_alias="ipV6VisibilityCacheEntries",
        ge=16,
        le=2000000,
    )
    model_config = ConfigDict(populate_by_name=True)


class LocalizedPolicyAssemblyItem(AssemblyItemBase):
    type: LocalizedPolicySupportedItemType
    definition_id: UUID = Field(serialization_alias="definitionId", validation_alias="definitionId")
    model_config = ConfigDict(populate_by_name=True)


class LocalizedPolicyDefinition(PolicyDefinition):
    assembly: List[LocalizedPolicyAssemblyItem]
    settings: LocalizedPolicySettings


class LocalizedPolicy(PolicyCreationPayload):
    policy_definition: LocalizedPolicyDefinition = Field(
        default=LocalizedPolicyDefinition(assembly=[], settings=LocalizedPolicySettings()),
        serialization_alias="policyDefinition",
        validation_alias="policyDefinition",
    )
    policy_type: str = Field(default="feature", serialization_alias="policyType", validation_alias="policyType")

    def _add_item(self, type: LocalizedPolicySupportedItemType, id: UUID) -> None:
        self.policy_definition.assembly.append(LocalizedPolicyAssemblyItem(type=type, definition_id=id))

    def add_qos_map(self, definition_id: UUID) -> None:
        self._add_item("qosMap", definition_id)

    def add_rewrite_rule(self, definition_id: UUID) -> None:
        self._add_item("rewriteRule", definition_id)

    def add_vpn_qos_map(self, definition_id: UUID) -> None:
        self._add_item("vpnQoSMap", definition_id)

    def add_access_control_list(self, definition_id: UUID) -> None:
        self._add_item("acl", definition_id)

    def add_access_control_list_ipv6(self, definition_id: UUID) -> None:
        self._add_item("aclv6", definition_id)

    def add_device_access_policy(self, definition_id: UUID) -> None:
        self._add_item("deviceAccessPolicy", definition_id)

    def add_device_access_policy_ipv6(self, definition_id: UUID) -> None:
        self._add_item("deviceAccessPolicyv6", definition_id)

    def add_route_policy(self, definition_id: UUID) -> None:
        self._add_item("vedgeRoute", definition_id)

    @field_validator("policy_definition", mode="before")
    @classmethod
    def try_parse(cls, policy_definition):
        # this is needed because GET /template/policy/vedge contains string in policyDefinition field
        # while POST /template/policy/vedge requires a regular object
        # it makes sense to reuse that model for both requests and present parsed data to the user
        if isinstance(policy_definition, str):
            return LocalizedPolicyDefinition.model_validate_json(policy_definition)
        return policy_definition


class LocalizedPolicyInfo(PolicyInfo, LocalizedPolicy):
    pass


class LocalizedPolicyEditResponse(BaseModel):
    master_templates_affected: List[str] = Field(
        default=[], serialization_alias="masterTemplatesAffected", validation_alias="masterTemplatesAffected"
    )


class LocalizedPolicyDeviceInfo(BaseModel):
    local_system_ip: IPvAnyAddress = Field(serialization_alias="local-system-ip", validation_alias="local-system-ip")
    host_name: str = Field(serialization_alias="host-name", validation_alias="host-name")
    site_id: Optional[str] = Field(None, serialization_alias="site-id", validation_alias="site-id")
    uuid: UUID
    layout_level: int = Field(serialization_alias="layoutLevel", validation_alias="layoutLevel")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

"""This modules defines APIEndpoints class which is used to define
vManage API endpoint handlers in declarative way.
Just create a sub-class and define endpoints using using included decorators: request, view, versions.
Method decorated with @request has no body, as decorator constructs and sends request.
>>> from catalystwan.endpoints import APIEndpoints, versions, view, request
>>> from catalystwan.utils.session_type import ProviderView
>>>
>>>
>>> class TenantBulkDeleteRequest(BaseModel):
>>>     password: str
>>>     tenant_id_list: List[str] = Field(alias="tenantIdList")
>>>
>>>
>>> class TenantTaskId(BaseModel):
>>>     id: str
>>>
>>>
>>> class TenantManagementAPI(APIEndpoints):
>>>     @versions(">=20.4")
>>>     @view({ProviderView})
>>>     @delete("/tenant/bulk/async")
>>>     def delete_tenant_async_bulk(self, payload: TenantBulkDeleteRequest) -> TenantTaskId:
>>>         ...
>>>
To send request instantiate API with logged ManagerSession:
>>> api = TenantManagementAPI(session)
>>> api.delete_tenant_async_bulk(TenantBulkDeleteRequest(password="", tenantIdList=["TNT00005"]))
"""
from __future__ import annotations

import json
import logging
from dataclasses import dataclass, fields
from enum import Enum
from inspect import _empty, isclass, signature
from io import BufferedReader
from string import Formatter
from typing import (
    Any,
    BinaryIO,
    ClassVar,
    Dict,
    Final,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    runtime_checkable,
)
from uuid import UUID

from packaging.specifiers import SpecifierSet  # type: ignore
from packaging.version import Version  # type: ignore
from pydantic import BaseModel
from typing_extensions import Annotated, get_args, get_origin

from catalystwan.abstractions import APIEndpointClient, APIEndpointClientResponse
from catalystwan.exceptions import APIEndpointError, APIRequestPayloadTypeError, APIVersionError, APIViewError
from catalystwan.typed_list import DataSequence
from catalystwan.utils.session_type import SessionType

BASE_PATH: Final[str] = "/dataservice"
T = TypeVar("T")
logger = logging.getLogger(__name__)


@runtime_checkable
class CustomPayloadType(Protocol):
    def prepared(self) -> PreparedPayload:
        ...


JSON = Union[str, int, float, bool, None, Dict[str, "JSON"], List["JSON"]]
ModelPayloadType = Union[BaseModel, Sequence[BaseModel]]
PayloadType = Union[None, JSON, str, bytes, dict, ModelPayloadType, CustomPayloadType]
ReturnType = Union[None, JSON, bytes, str, dict, BaseModel, DataSequence[BaseModel]]
RequestParamsType = Union[Dict[str, str], BaseModel]


@dataclass
class TypeSpecifier:
    """Holds type information extracted from signature. Common for payload and return values. Used for documentation"""

    present: bool
    sequence_type: Optional[type] = None
    payload_type: Optional[type] = None
    payload_union_model_types: Optional[Sequence[type]] = None
    is_json: bool = False  # JSON is treated specially as it is type-alias / <typing special form>
    is_optional: bool = False

    @classmethod
    def not_present(cls) -> TypeSpecifier:
        return TypeSpecifier(present=False)

    @classmethod
    def none_type(cls) -> TypeSpecifier:
        return TypeSpecifier(present=True)

    @classmethod
    def json(cls) -> TypeSpecifier:
        return TypeSpecifier(present=True, is_json=True)

    @classmethod
    def model_union(cls, models: Sequence[type]) -> TypeSpecifier:
        return TypeSpecifier(present=True, payload_union_model_types=models)


@dataclass
class APIEndpointRequestMeta:
    """Holds data for endpoints exctracted during decorating. Used for documentation"""

    func: Any
    http_request: str
    payload_spec: TypeSpecifier
    return_spec: TypeSpecifier


@dataclass(frozen=True)
class PreparedPayload:
    """Holds data prepared for sending in request"""

    data: Union[Dict, str, bytes, None] = None
    headers: Optional[Mapping[str, Any]] = None
    files: Optional[Dict[str, Tuple[str, BufferedReader]]] = None

    def asdict(self) -> Dict[str, Any]:
        """Needed because python builtin does deepcopy of objects (it is impossible for BufferedReader)"""
        result = {}
        for f in fields(self):
            if value := getattr(self, f.name):
                result[f.name] = value
        return result


def dict_values_to_str(field_names: Set[str], kwargs: Dict[str, Any]) -> Dict[str, str]:
    # this is to keep compatiblity and have seme behavior for (str, Enum) mixin after 3.11 for url formatting
    result: Dict[str, str] = {}
    for field_name in field_names:
        field_value = kwargs.get(field_name)
        if isinstance(field_value, Enum):
            result[field_name] = str(field_value.value)
        else:
            result[field_name] = str(field_value)
    return result


class APIEndpoints:
    """
    Class to be used as base for all API endpoints.
    Injects BASE_PATH url prefix as it is common for all known vManage API endpoints.
    Introduces special keyword argument 'payload' in request call and handles sending of such object.
    """

    @classmethod
    def _prepare_payload(cls, payload: PayloadType, force_json: bool = False) -> PreparedPayload:
        """Helper method to prepare data for sending based on type"""
        if force_json or isinstance(payload, dict):
            return PreparedPayload(data=json.dumps(payload), headers={"content-type": "application/json"})
        if isinstance(payload, (str, bytes)):
            return PreparedPayload(data=payload)
        elif isinstance(payload, (BaseModel)):
            return cls._prepare_basemodel_payload(payload)
        elif isinstance(payload, Sequence) and not isinstance(payload, (str, bytes)):
            return cls._prepare_sequence_payload(payload)  # type: ignore[arg-type]
            # offender is List[JSON] which is also a Sequence can be ignored as long as force_json is passed correctly
        elif isinstance(payload, CustomPayloadType):
            return payload.prepared()
        else:
            raise APIRequestPayloadTypeError(payload)

    @classmethod
    def _prepare_basemodel_payload(cls, payload: BaseModel) -> PreparedPayload:
        """Helper method to prepare BaseModel instance for sending"""
        return PreparedPayload(
            data=payload.model_dump_json(exclude_none=True, by_alias=True), headers={"content-type": "application/json"}
        )

    @classmethod
    def _prepare_sequence_payload(cls, payload: Iterable[BaseModel]) -> PreparedPayload:
        """Helper method to prepare sequences for sending"""
        items = []
        for item in payload:
            items.append(item.model_dump(exclude_none=True, by_alias=True))
        data = json.dumps(items)
        return PreparedPayload(data=data, headers={"content-type": "application/json"})

    @classmethod
    def _prepare_params(cls, params: RequestParamsType) -> Dict[str, Any]:
        """Helper method to prepare params for sending"""
        if isinstance(params, BaseModel):
            return params.model_dump(exclude_none=True, by_alias=True)
        return params

    def __init__(self, client: APIEndpointClient):
        self._client = client
        self._basepath = BASE_PATH

    def _request(
        self,
        method: str,
        url: str,
        payload: Optional[ModelPayloadType] = None,
        params: Optional[RequestParamsType] = None,
        force_json_payload: bool = False,
        **kwargs,
    ) -> APIEndpointClientResponse:
        """Prepares and sends request using client protocol"""
        _kwargs = dict(kwargs)
        if payload is not None:
            _kwargs.update(self._prepare_payload(payload, force_json_payload).asdict())
        if params is not None:
            _kwargs.update({"params": self._prepare_params(params)})
        return self._client.request(method, self._basepath + url, **_kwargs)

    @property
    def _api_version(self) -> Optional[Version]:
        return self._client.api_version

    @property
    def _session_type(self) -> Optional[SessionType]:
        return self._client.session_type


class APIEndpointsDecorator:
    @classmethod
    def get_check_instance(cls, _self, *args, **kwargs) -> APIEndpoints:
        """Gets wrapped function instance (first argument)"""
        if not isinstance(_self, APIEndpoints):
            raise APIEndpointError("Only APIEndpointsDecorator instance methods can be annotated with @{cls} decorator")
        return _self


class versions(APIEndpointsDecorator):
    """
    Decorator to annotate endpoints with supported versions.
    Logs warning or raises exception when incompatibility found during runtime.
    """

    versions_lookup: ClassVar[
        Dict[str, SpecifierSet]
    ] = {}  # maps decorated method instance to it's supported verisions

    def __init__(self, supported_versions: str, raises: bool = False):
        self.supported_versions = SpecifierSet(supported_versions)
        self.raises = raises

    def __call__(self, func):
        original_func = getattr(func, "_ofunc", func)  # grab original function
        self.versions_lookup[original_func.__qualname__] = self.supported_versions

        def wrapper(*args, **kwargs):
            """Executes each time decorated method is called"""
            _self = self.get_check_instance(*args, **kwargs)  # _self refers to APIEndpoints instance
            current = _self._api_version
            supported = self.supported_versions
            if current and current not in supported:
                if self.raises:
                    raise APIVersionError(func, supported, current)
                else:
                    logger.warning(
                        f"vManage runs: {current} but {func.__qualname__} only supported for API versions: {supported}"
                    )
            return func(*args, **kwargs)

        wrapper._ofunc = original_func  # provide original function to next decorator in chain
        return wrapper


class view(APIEndpointsDecorator):
    """
    Decorator to annotate endpoints with session type (view) restriction
    Logs warning or raises exception when incompatibility found during runtime.
    """

    view_lookup: ClassVar[Dict[str, Set[SessionType]]] = {}  # maps decorated method instance to it's allowed sessions

    def __init__(self, allowed_session_types: Set[SessionType], raises: bool = False):
        self.allowed_session_types = allowed_session_types
        self.raises = raises

    def __call__(self, func):
        original_func = getattr(func, "_ofunc", func)  # grab original function
        self.view_lookup[original_func.__qualname__] = self.allowed_session_types

        def wrapper(*args, **kwargs):
            """Executes each time decorated method is called"""
            _self = self.get_check_instance(*args, **kwargs)  # _self refers to APIEndpoints instance
            current = _self._session_type
            allowed = self.allowed_session_types
            if current and current not in allowed:
                if self.raises:
                    raise APIViewError(func, allowed, current)
                else:
                    logger.warning(
                        f"Current view is: {current} but {func.__qualname__} only allowed for views: {allowed}"
                    )
            return func(*args, **kwargs)

        wrapper._ofunc = original_func  # provide original function to next decorator in chain
        return wrapper


class request(APIEndpointsDecorator):
    """
    Decorator to annotate endpoints with HTTP method, URL and optionally json key from which
    modelled data will be parsed (usually "data", but defaults to whole json payload).
    Additional kwargs can be injected which will be passed to request method (eg. custom headers)

    Decorated method parameters and return type annotations are checked:

        Parameters:

            "payload": argument with that name is used to send data in request
            usually it is user defined subclass of: pydantic.BaseModel
            other types are also supported please check: catalystwan.endpoints.PayloadType

            "params": argument with that name is used to generate url query string
            supports Dict[str, str] and pydantic.BaseModel

            other parameter must be strings and are used to format url.

        Return Type:

            supports types defined in: catalystwan.endpoints.ReturnType

    Raises:
        APIEndpointError: when decorated method has unsupported parameters or response type
    """

    forbidden_url_field_names = {"self", "payload", "params"}
    request_lookup: ClassVar[
        Dict[str, APIEndpointRequestMeta]
    ] = {}  # maps decorated method instance to it's meta information

    def __init__(self, http_method: str, url: str, resp_json_key: Optional[str] = None, **kwargs):
        self.http_method = http_method
        formatter = Formatter()
        url_field_names = {item[1] for item in formatter.parse(url) if item[1] is not None}
        if self.forbidden_url_field_names & url_field_names:
            APIEndpointError(f"One of forbidden fields names: {self.forbidden_url_field_names} found in url: {url}")
        self.url = url
        self.url_field_names = url_field_names
        self.resp_json_key = resp_json_key
        self.return_spec = TypeSpecifier.not_present()
        self.payload_spec = TypeSpecifier.not_present()
        self.kwargs = kwargs

    def specify_return_type(self) -> TypeSpecifier:
        """Specifies return type based on decorated method signature annotations.
        Does basic checking of annotated types so problems can be detected early.

        Raises:
            APIEndpointError: when signature contains unexpected return annotation

        Returns:
            TypeSpecifier: Specification of return type
        """
        annotation = self.sig.return_annotation
        if annotation == JSON:
            return TypeSpecifier.json()
        if annotation is None:
            return TypeSpecifier.none_type()
        if annotation == _empty:
            raise APIEndpointError(
                "APIEndpoint methods decorated with @request must specify return type, "
                "use None annotation if function does not return any value"
            )
        if (type_origin := get_origin(annotation)) and isclass(type_origin) and issubclass(type_origin, DataSequence):
            if (
                (type_args := get_args(annotation))
                and (len(type_args) == 1)
                and isclass(type_args[0])
                and issubclass(type_args[0], BaseModel)
            ):
                return TypeSpecifier(True, DataSequence, type_args[0])
            raise APIEndpointError(f"Expected: {ReturnType} but return type {annotation}")
        elif isclass(annotation):
            try:
                if issubclass(annotation, (bytes, str, dict, BinaryIO, BaseModel)):
                    return TypeSpecifier(True, None, annotation)
                raise APIEndpointError(f"Expected: {ReturnType} but return type {annotation}")
            except TypeError:
                raise APIEndpointError(f"Expected: {ReturnType} but return type {annotation}")
        raise APIEndpointError(f"Expected: {ReturnType} but return type {annotation}")

    def specify_payload_type(self) -> TypeSpecifier:
        """Specifies payload type based on decorated method signature annotations.
        Does basic checking of annotated types for 'payload' so problems can be detected early.

        Raises:
            APIEndpointError: when signature contains unexpected payload annotation

        Returns:
            TypeSpecifier: Specification of payload type
        """
        payload_param = self.sig.parameters.get("payload")
        if not payload_param:
            return TypeSpecifier.not_present()

        annotation = payload_param.annotation
        is_optional = False

        # Check if JSON
        if annotation == JSON:
            return TypeSpecifier.json()

        # Check if Optional (flag and replace original annotation with optional type if exactly one is present)
        if type_origin := get_origin(annotation):
            type_args = get_args(annotation)
            if type_origin == Union and (type(None) in type_args):
                optional_type_args = tuple(arg for arg in get_args(annotation) if not arg == type(None))  # noqa: E721
                # flake suggest using isinstance(arg, type(None)) above, but it doesn't match NoneType
                if len(optional_type_args) == 1:
                    is_optional = True
                    annotation = optional_type_args[0]

        # Check if regular class
        if isclass(annotation):
            if issubclass(annotation, (bytes, str, dict, BinaryIO, BaseModel, CustomPayloadType)):
                return TypeSpecifier(True, None, annotation, None, False, is_optional)
            else:
                raise APIEndpointError(f"'payload' param must be annotated with supported type: {PayloadType}")

        # Check for accepted alias types like List[...] and Union[...]
        elif type_origin := get_origin(annotation):
            # Check if Sequence[PayloadModelType] like List or DataSequence
            if isclass(type_origin) and issubclass(type_origin, Sequence):
                if (
                    (type_args := get_args(annotation))
                    and (len(type_args) == 1)
                    and isclass(type_args[0])
                    and issubclass(type_args[0], BaseModel)
                ):
                    return TypeSpecifier(True, type_origin, type_args[0], None, False, is_optional)
            # Check if Annnotated[Union[PayloadModelType, ...]], only unions of pydantic models allowed
            elif type_origin == Annotated:
                if annotated_origin := get_args(annotation):
                    if (len(annotated_origin) >= 1) and get_origin(annotated_origin[0]) == Union:
                        if (
                            (type_args := get_args(annotated_origin[0]))
                            and all(isclass(t) for t in type_args)
                            and all(issubclass(t, BaseModel) for t in type_args)
                        ):
                            return TypeSpecifier.model_union(models=list(type_args))
            # Check if Union[PayloadModelType, ...], only unions of pydantic models allowed
            elif type_origin == Union:
                if (
                    (type_args := get_args(annotation))
                    and all(isclass(t) for t in type_args)
                    and all(issubclass(t, BaseModel) for t in type_args)
                ):
                    return TypeSpecifier.model_union(models=list(type_args))
            raise APIEndpointError(f"Expected: {PayloadType} but found payload {annotation}")
        else:
            raise APIEndpointError(f"Expected: {PayloadType} but found payload {annotation}")

    def check_params(self):
        """Checks params in decorated method definition

        Raises:
            APIEndpointError: when decorated params not matching specification
        """
        parameters = self.sig.parameters

        if params_param := parameters.get("params"):
            if not (isclass(params_param.annotation) and issubclass(params_param.annotation, (BaseModel, Dict))):
                raise APIEndpointError(f"'params' param must be annotated with supported type: {RequestParamsType}")

        general_purpose_arg_names = {
            key for key in self.sig.parameters.keys() if key not in self.forbidden_url_field_names
        }
        if missing := self.url_field_names.difference(general_purpose_arg_names):
            raise APIEndpointError(f"Missing parameters: {missing} to format url: {self.url}")

        for parameter in [parameters.get(name) for name in self.url_field_names]:
            # Check if 'params' is type of str, UUID or LIteral
            if not (isclass(parameter.annotation) and issubclass(parameter.annotation, (str, UUID))):
                if not get_origin(parameter.annotation) == Literal:
                    raise APIEndpointError(
                        f"Parameter {parameter} used for url formatting must be 'str', UUID or Literal sub-type"
                    )

                elif p_args := get_args(parameter.annotation):
                    # Check if all 'params' Literal values are str
                    if not all((isinstance(arg, str) for arg in p_args)):
                        raise APIEndpointError(
                            f"Literal values for parameter {parameter} used for url formatting must be 'str'"
                        )

        no_purpose_params = {
            parameters.get(name) for name in general_purpose_arg_names.difference(self.url_field_names)
        }
        if no_purpose_params:
            raise APIEndpointError(
                f"Parameters {no_purpose_params} are not used as "
                "request payload, request params nor to format url string!"
                "remove unused parameter or fix by changing argument to purposeful name 'payload' or 'params'"
            )

    def merge_args(self, positional_args: Tuple, keyword_args: Dict[str, Any]) -> Dict[str, Any]:
        """Merges decorated method args and kwargs into one dictionary.
        This is needed to identify all decorated method arguments by name inside wrapper body.
        We can learn all arguments names from signature.

        Returns: Dict[str, Any]: all passed args as keyword arguments (excluding "self")
        """
        all_args_names = [key for key in self.sig.parameters.keys()]
        all_args_dict = dict(self.defaults)
        all_args_dict.update(dict(zip(all_args_names, positional_args)))
        all_args_dict.update(keyword_args)
        all_args_dict.pop("self", None)
        return all_args_dict

    def __call__(self, func):
        original_func = getattr(func, "_ofunc", func)  # grab original function
        self.sig = signature(original_func)
        self.defaults = {
            key: value.default for (key, value) in self.sig.parameters.items() if value.default is not _empty
        }
        self.return_spec = self.specify_return_type()
        self.payload_spec = self.specify_payload_type()
        self.check_params()
        self.request_lookup[original_func.__qualname__] = APIEndpointRequestMeta(
            func=original_func,
            http_request=f"{self.http_method} {self.url}",
            payload_spec=self.payload_spec,
            return_spec=self.return_spec,
        )

        def wrapper(*args, **kwargs):
            """Executes each time decorated method is called"""
            _self = self.get_check_instance(*args, **kwargs)  # _self refers to APIEndpoints instance
            _kwargs = self.merge_args(args, kwargs)
            payload = _kwargs.get("payload")
            params = _kwargs.get("params")
            url_kwargs = dict_values_to_str(self.url_field_names, _kwargs)
            formatted_url = self.url.format_map(url_kwargs)
            response = _self._request(
                self.http_method,
                formatted_url,
                payload=payload,
                force_json_payload=self.payload_spec.is_json,
                params=params,
                **self.kwargs,
            )
            if self.return_spec.present:
                if self.return_spec.is_json:
                    full_json = response.json()
                    if self.resp_json_key is not None:
                        if isinstance(full_json, dict):
                            return full_json.get(self.resp_json_key)
                        else:
                            raise TypeError(f"Expected dictionary as json payload but found: {type(full_json)}")
                    return full_json
                if self.return_spec.payload_type is None:
                    pass
                elif issubclass(self.return_spec.payload_type, BaseModel):
                    if self.return_spec.sequence_type == DataSequence:
                        return response.dataseq(
                            cls=self.return_spec.payload_type,
                            sourcekey=self.resp_json_key,
                            validate=_self._client.validate_responses,
                        )
                    else:
                        return response.dataobj(
                            self.return_spec.payload_type, self.resp_json_key, validate=_self._client.validate_responses
                        )
                elif issubclass(self.return_spec.payload_type, str):
                    return response.text
                elif issubclass(self.return_spec.payload_type, bytes):
                    return response.content
                elif issubclass(self.return_spec.payload_type, dict):
                    return response.json()

        wrapper._ofunc = original_func  # provide original function to next decorator in chain
        return wrapper


class get(request):
    def __init__(self, url: str, resp_json_key: Optional[str] = None, **kwargs):
        super().__init__("GET", url, resp_json_key, **kwargs)


class put(request):
    def __init__(self, url: str, resp_json_key: Optional[str] = None, **kwargs):
        super().__init__("PUT", url, resp_json_key, **kwargs)


class post(request):
    def __init__(self, url: str, resp_json_key: Optional[str] = None, **kwargs):
        super().__init__("POST", url, resp_json_key, **kwargs)


class delete(request):
    def __init__(self, url: str, resp_json_key: Optional[str] = None, **kwargs):
        super().__init__("DELETE", url, resp_json_key, **kwargs)



# Copyright 2022 Cisco Systems, Inc. and its affiliates

import logging
import logging.config
import multiprocessing
from functools import lru_cache, wraps
from importlib import metadata
from importlib.machinery import PathFinder
from os import environ
from pathlib import Path
from traceback import FrameSummary, StackSummary, extract_stack
from typing import Callable, Final, List, Optional

import urllib3

USER_AGENT = f"{__package__}/{metadata.version(__package__)}"


def with_proc_info_header(method: Callable[..., str]) -> Callable[..., str]:
    """
    Adds process ID and external caller information before first line of returned string
    """

    @wraps(method)
    def wrapper(*args, **kwargs) -> str:
        wrapped = method(*args, **kwargs)
        header = f"{multiprocessing.current_process()}"
        if frame_summary := get_first_external_stack_frame(extract_stack()):
            fname, line_no, function, _ = frame_summary
            header += " %s:%d %s(...)" % (fname, line_no, function)
        header += "\n"
        return header + wrapped

    return wrapper


def get_first_external_stack_frame(stack: StackSummary) -> Optional[FrameSummary]:
    """
    Get the first python frame
    on the stack before entering catalystwan module
    """
    if len(stack) < 1:
        return None
    for index, frame in enumerate(stack):
        if is_file_in_package(frame.filename):
            break
    if index == 0:
        return None
    return stack[index - 1]


@lru_cache()
def is_file_in_package(fname: str) -> bool:
    """
    Checks if filepath given by string
    is part of catalystwan source code
    """
    return Path(fname) in pkg_src_list


def list_package_sources() -> List[Path]:
    """
    Creates a list containing paths to all python source files
    for current package
    """
    pkg_srcs: List[Path] = []
    if pkg_spec := PathFinder.find_spec(__package__):
        if pkg_origin := pkg_spec.origin:
            pkg_srcs = list(Path(pkg_origin).parent.glob("**/*.py"))
    return pkg_srcs


LOGGING_CONF_DIR: Final[str] = str(Path(__file__).parents[0] / "logging.conf")
__version__ = metadata.version(__package__)
pkg_src_list = list_package_sources()


if environ.get("catalystwan_devel") is not None:
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    catalystwan_logger = logging.getLogger(__name__)
    logging.config.fileConfig(LOGGING_CONF_DIR, disable_existing_loggers=False)
    catalystwan_logger.debug(f"catalystwan {__version__}")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import datetime
from typing import List, Optional, Protocol
from uuid import UUID

from pydantic import BaseModel, Field

from catalystwan.models.policy import AnyPolicyList
from catalystwan.typed_list import DataSequence


class InfoTag(BaseModel):
    info_tag: Optional[str] = Field("", alias="infoTag")


class PolicyListId(BaseModel):
    list_id: UUID = Field(alias="listId")


class PolicyListInfo(PolicyListId, InfoTag):
    last_updated: datetime.datetime = Field(alias="lastUpdated")
    owner: str
    read_only: bool = Field(alias="readOnly")
    version: str
    reference_count: int = Field(alias="referenceCount")
    references: List
    is_activated_by_vsmart: Optional[bool] = Field(None, alias="isActivatedByVsmart")


class PolicyListPreview(BaseModel):
    preview: str


class PolicyListEndpoints(Protocol):
    def create_policy_list(self, payload: AnyPolicyList) -> PolicyListId:
        ...

    def delete_policy_list(self, id: UUID) -> None:
        ...

    def edit_policy_list(self, id: UUID, payload: AnyPolicyList) -> None:
        ...

    def get_lists_by_id(self, id: UUID) -> PolicyListInfo:
        ...

    def get_policy_lists(self) -> DataSequence[PolicyListInfo]:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Network, IPv6Network
from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing_extensions import Annotated

from catalystwan.models.common import check_fields_exclusive
from catalystwan.models.policy.policy_definition import PolicyDefinitionBase, Reference, VariableName


class RuleBase(BaseModel):
    rule: str = ""
    order: str = ""
    action: str = "permit"
    source_security_group: Optional[Reference] = Field(
        default=None, serialization_alias="sourceSecurityGroup", validation_alias="sourceSecurityGroup"
    )
    destination_security_group: Optional[Reference] = Field(
        default=None, serialization_alias="destinationSecurityGroup", validation_alias="destinationSecurityGroup"
    )
    protocol: Optional[str] = None
    protocol_name: Optional[str] = Field(
        default=None, serialization_alias="protocolName", validation_alias="protocolName"
    )
    protocol_name_list: Optional[Reference] = Field(
        default=None, serialization_alias="protocolNameList", validation_alias="protocolNameList"
    )
    model_config = ConfigDict(populate_by_name=True)

    @model_validator(mode="after")
    def check_exclusive_fields(self):
        check_fields_exclusive(self.__dict__, {"protocol", "protocol_name", "protocol_name_list"}, False)
        return self


class IPv4Rule(RuleBase):
    sequence_ip_type: Literal["ipv4"] = Field(
        default="ipv4", serialization_alias="sequenceIpType", validation_alias="sequenceIpType"
    )
    source_ip: Union[IPv4Network, VariableName, None] = Field(
        default=None, serialization_alias="sourceIP", validation_alias="sourceIP"
    )
    source_data_prefix_list: Optional[Reference] = Field(
        default=None, serialization_alias="sourceDataPrefixList", validation_alias="sourceDataPrefixList"
    )
    source_fqdn: Optional[str] = Field(default=None, serialization_alias="sourceFqdn", validation_alias="sourceFqdn")
    source_fqdn_list: Optional[Reference] = Field(
        default=None, serialization_alias="sourceFqdnList", validation_alias="sourceFqdnList"
    )
    source_geo_location: Optional[str] = Field(
        default=None, serialization_alias="sourceGeoLocation", validation_alias="sourceGeoLocation"
    )
    source_geo_location_list: Optional[Reference] = Field(
        default=None, serialization_alias="sourceGeoLocationList", validation_alias="sourceGeoLocationList"
    )
    source_port: Optional[str] = Field(default=None, serialization_alias="sourcePort", validation_alias="sourcePort")
    source_port_list: Optional[Reference] = Field(
        default=None, serialization_alias="sourcePortList", validation_alias="sourcePortList"
    )
    destination_ip: Union[IPv4Network, VariableName, None] = Field(
        default=None, serialization_alias="destinationIP", validation_alias="destinationIP"
    )
    destination_data_prefix_list: Optional[Reference] = Field(
        default=None, serialization_alias="destinationDataPrefixList", validation_alias="destinationDataPrefixList"
    )
    destination_fqdn: Optional[str] = Field(
        default=None, serialization_alias="destinationFqdn", validation_alias="destinationFqdn"
    )
    destination_fqdn_list: Optional[Reference] = Field(
        default=None, serialization_alias="destinationFqdnList", validation_alias="destinationFqdnList"
    )
    destination_geo_location: Optional[str] = Field(
        default=None, serialization_alias="destinationGeoLocation", validation_alias="destinationGeoLocation"
    )
    destination_geo_location_list: Optional[Reference] = Field(
        default=None, serialization_alias="destinationGeoLocationList", validation_alias="destinationGeoLocationList"
    )
    destination_port: Optional[str] = Field(
        default=None, serialization_alias="destinationPort", validation_alias="destinationPort"
    )
    destination_port_list: Optional[Reference] = Field(
        default=None, serialization_alias="destinationPortList", validation_alias="destinationPortList"
    )
    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    @model_validator(mode="after")
    def check_exclusive_fields(self):
        check_fields_exclusive(self.__dict__, {"source_security_group", "source_ip", "source_data_prefix_list"}, False)
        check_fields_exclusive(self.__dict__, {"source_security_group", "source_fqdn", "source_fqdn_list"}, False)
        check_fields_exclusive(
            self.__dict__, {"source_security_group", "source_geo_location", "source_geo_location_list"}, False
        )
        check_fields_exclusive(self.__dict__, {"source_security_group", "source_port", "source_port_list"}, False)
        check_fields_exclusive(
            self.__dict__, {"destination_security_group", "destination_ip", "destination_data_prefix_list"}, False
        )
        check_fields_exclusive(
            self.__dict__, {"destination_security_group", "destination_fqdn", "destination_fqdn_list"}, False
        )
        check_fields_exclusive(
            self.__dict__,
            {"destination_security_group", "destination_geo_location", "destination_geo_location_list"},
            False,
        )
        check_fields_exclusive(
            self.__dict__, {"destination_security_group", "destination_port", "destination_port_list"}, False
        )
        return self


class IPv6Rule(RuleBase):
    sequence_ip_type: Literal["ipv6"] = Field(
        default="ipv6", serialization_alias="sequenceIpType", validation_alias="sequenceIpType"
    )
    source_ipv6: Union[IPv6Network, VariableName, None] = Field(
        default=None, serialization_alias="sourceIPV6", validation_alias="sourceIPV6"
    )
    source_ipv6_data_prefix_list: Optional[Reference] = Field(
        default=None, serialization_alias="sourceIPV6DataPrefixList", validation_alias="sourceIPV6DataPrefixList"
    )
    destination_ipv6: Union[IPv6Network, VariableName, None] = Field(
        default=None, serialization_alias="destinationIPV6", validation_alias="destinationIPV6"
    )
    destination_ipv6_data_prefix_list: Optional[Reference] = Field(
        default=None,
        serialization_alias="destinationIPV6DataPrefixList",
        validation_alias="destinationIPV6DataPrefixList",
    )
    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    @model_validator(mode="after")
    def check_exclusive_fields(self):
        check_fields_exclusive(
            self.__dict__, {"source_security_group", "source_ipv6", "source_ipv6_data_prefix_list"}, False
        )
        check_fields_exclusive(
            self.__dict__,
            {"destination_security_group", "destination_ipv6", "destination_ipv6_data_prefix_list"},
            False,
        )
        return self


Rule = Annotated[Union[IPv4Rule, IPv6Rule], Field(discriminator="sequence_ip_type")]


class RuleSetDefinition(BaseModel):
    rules: List[Rule] = []


class RuleSet(PolicyDefinitionBase):
    type: Literal["ruleSet"] = "ruleSet"
    definition: RuleSetDefinition = RuleSetDefinition()

    def _enumerate_rules(self, from_index: int = 0) -> None:
        """Updates rules entries with appropriate order and rule values.

        Args:
            from_index (int, optional): Only rules after that index in table will be updated. Defaults to 0.
        """
        start_index = from_index
        rule_count = len(self.definition.rules)
        if from_index < 0:
            start_index = rule_count - start_index
        for i in range(start_index, rule_count):
            order = str(i + 1)
            self.definition.rules[i].order = order
            self.definition.rules[i].rule = f"R{order}"

    def pop(self, index: int = -1) -> None:
        """Removes a rule at given index, consecutive rules will be enumarated again.

        Args:
            index (int, optional): Defaults to -1.
        """
        self.definition.rules.pop(index)
        self._enumerate_rules(index)

    def add(self, rule: Rule) -> None:
        """Adds new rule as last in table, order and rule fields will be autogenerated.

        Args:
            rule (Rule)
        """
        insert_index = len(self.definition.rules)
        self.definition.rules.append(rule)
        self._enumerate_rules(insert_index)

    def add_ipv4_rule(
        self,
        action: str = "permit",
        source_security_group_id: Optional[UUID] = None,
        source_ip: Optional[IPv4Network] = None,
        source_ip_variable: Optional[str] = None,
        source_data_prefix_list_id: Optional[UUID] = None,
        source_fqdn: Optional[str] = None,
        source_fqdn_list_id: Optional[UUID] = None,
        source_geo_location: Optional[str] = None,
        source_geo_location_list_id: Optional[UUID] = None,
        source_port: Optional[str] = None,
        source_port_list_id: Optional[UUID] = None,
        destination_security_group_id: Optional[UUID] = None,
        destination_ip: Optional[IPv4Network] = None,
        destination_ip_variable: Optional[str] = None,
        destination_data_prefix_list_id: Optional[UUID] = None,
        destination_fqdn: Optional[str] = None,
        destination_fqdn_list_id: Optional[UUID] = None,
        destination_geo_location: Optional[str] = None,
        destination_geo_location_list_id: Optional[UUID] = None,
        destination_port: Optional[str] = None,
        destination_port_list_id: Optional[UUID] = None,
        protocols: Optional[List[int]] = None,
        protocol_names: Optional[str] = None,
        protocol_name_list_id: Optional[UUID] = None,
    ) -> None:
        if source_ip is not None and source_ip_variable is not None:
            raise ValueError("Source IP and variable name cannot be set at the same time")
        if destination_ip is not None and destination_ip_variable is not None:
            raise ValueError("Destination IP and variable name cannot be set at the same time")
        ipv4_rule = IPv4Rule(
            action=action,
            source_security_group=Reference(ref=source_security_group_id) if source_security_group_id else None,
            source_ip=source_ip or (VariableName(vip_variable_name=source_ip_variable) if source_ip_variable else None),
            source_data_prefix_list=Reference(ref=source_data_prefix_list_id) if source_data_prefix_list_id else None,
            source_fqdn=source_fqdn,
            source_fqdn_list=Reference(ref=source_fqdn_list_id) if source_fqdn_list_id else None,
            source_geo_location=source_geo_location,
            source_geo_location_list=Reference(ref=source_geo_location_list_id)
            if source_geo_location_list_id
            else None,
            source_port=source_port,
            source_port_list=Reference(ref=source_port_list_id) if source_port_list_id else None,
            destination_security_group=Reference(ref=destination_security_group_id)
            if destination_security_group_id
            else None,
            destination_ip=destination_ip
            or (VariableName(vip_variable_name=destination_ip_variable) if destination_ip_variable else None),
            destination_data_prefix_list=Reference(ref=destination_data_prefix_list_id)
            if destination_data_prefix_list_id
            else None,
            destination_fqdn=destination_fqdn,
            destination_fqdn_list=Reference(ref=destination_fqdn_list_id) if destination_fqdn_list_id else None,
            destination_geo_location=destination_geo_location,
            destination_geo_location_list=Reference(ref=destination_geo_location_list_id)
            if destination_geo_location_list_id
            else None,
            destination_port=destination_port,
            destination_port_list=Reference(ref=destination_port_list_id) if destination_port_list_id else None,
            protocol=" ".join(str(p) for p in protocols) if protocols else None,
            protocol_name=" ".join(p for p in protocol_names) if protocol_names else None,
            protocol_name_list=Reference(ref=protocol_name_list_id) if protocol_name_list_id else None,
        )
        self.add(ipv4_rule)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator


class Server(FeatureTemplateValidator):
    model_config = ConfigDict(populate_by_name=True)

    name: str
    key: Optional[int] = None
    vpn: Optional[int] = None
    version: Optional[int] = None
    source_interface: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "source-interface"})
    prefer: Optional[BoolStr] = None


class Authentication(FeatureTemplateValidator):
    model_config = ConfigDict(populate_by_name=True)

    number: int
    md5: str


class CiscoNTPModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    server: List[Server] = Field(default=[])
    authentication: Optional[List[Authentication]] = Field(default=None, json_schema_extra={"data_path": ["keys"]})
    trusted: Optional[List[int]] = Field(default=None, json_schema_extra={"data_path": ["keys"]})
    enable: Optional[BoolStr] = Field(default=None, json_schema_extra={"data_path": ["master"]})
    stratum: Optional[int] = Field(default=None, json_schema_extra={"data_path": ["master"]})
    source: Optional[str] = Field(default=None, json_schema_extra={"data_path": ["master"]})

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_ntp"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, CustomPayloadType, PreparedPayload, delete, get, post, versions
from catalystwan.typed_list import DataSequence
from catalystwan.utils.personality import Personality


class UnlockDeviceDetail(BaseModel):
    device_id: str = Field(validation_alias="deviceId", serialization_alias="deviceId")
    device_ip: str = Field(validation_alias="deviceIP", serialization_alias="deviceIP")


class DeviceUnlockPayload(BaseModel):
    device_type: str = Field(validation_alias="deviceType", serialization_alias="deviceType")
    devices: List[UnlockDeviceDetail]


class DeviceUnlockResponse(BaseModel):
    parentTaskId: str


Protocol = Literal["DTLS", "TLS"]


class DeviceCreationPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_ip: str = Field(alias="deviceIP")
    generate_csr: bool = Field(alias="generateCSR")
    password: str
    personality: Personality
    port: Optional[str] = Field(default=None)
    protocol: Protocol = Field(default="DTLS")
    username: str


class DeviceDeletionResponse(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    local_delete_from_db: Optional[bool] = Field(
        default=None, validation_alias="localDeleteFromDB", serialization_alias="localDeleteFromDB"
    )
    id: Optional[str] = Field(default=None)
    status: Optional[str] = Field(default=None)


DeviceCategory = Literal["controllers", "vedges"]


class DeviceDetailsResponse(BaseModel):
    # Field "model_sku" has conflict with protected namespace "model_"
    model_config = ConfigDict(populate_by_name=True, protected_namespaces=())

    device_type: Optional[str] = Field(default=None, validation_alias="deviceType", serialization_alias="deviceType")
    serial_number: Optional[str] = Field(
        default=None, validation_alias="serialNumber", serialization_alias="serialNumber"
    )
    uuid: Optional[str] = None
    management_system_ip: Optional[str] = Field(
        default=None, validation_alias="managementSystemIP", serialization_alias="managementSystemIP"
    )
    chasis_number: Optional[str] = Field(
        default=None, validation_alias="chasisNumber", serialization_alias="chasisNumber"
    )
    config_operation_mode: Optional[str] = Field(
        default=None, validation_alias="configOperationMode", serialization_alias="configOperationMode"
    )
    device_model: Optional[str] = Field(default=None, validation_alias="deviceModel", serialization_alias="deviceModel")
    device_state: Optional[str] = Field(default=None, validation_alias="deviceState", serialization_alias="deviceState")
    validity: Optional[str] = None
    platform_family: Optional[str] = Field(
        default=None, validation_alias="platformFamily", serialization_alias="platformFamily"
    )
    username: Optional[str] = None
    device_csr: Optional[str] = Field(default=None, validation_alias="deviceCSR", serialization_alias="deviceCSR")
    device_csr_common_name: Optional[str] = Field(
        default=None, validation_alias="deviceCSRCommonName", serialization_alias="deviceCSRCommonName"
    )
    root_cert_hash: Optional[str] = Field(
        default=None, validation_alias="rootCertHash", serialization_alias="rootCertHash"
    )
    csr: Optional[str] = Field(default=None, validation_alias="CSR", serialization_alias="CSR")
    csr_detail: Optional[str] = Field(default=None, validation_alias="CSRDetail", serialization_alias="CSRDetail")
    state: Optional[str] = None
    global_state: Optional[str] = Field(default=None, validation_alias="globalState", serialization_alias="globalState")
    valid: Optional[str] = None
    request_token_id: Optional[str] = Field(
        default=None, validation_alias="requestTokenID", serialization_alias="requestTokenID"
    )
    expiration_date: Optional[str] = Field(
        default=None, validation_alias="expirationDate", serialization_alias="expirationDate"
    )
    expiration_date_long: Optional[int] = Field(
        default=None, validation_alias="expirationDateLong", serialization_alias="expirationDateLong"
    )
    device_ip: Optional[str] = Field(default=None, validation_alias="deviceIP", serialization_alias="deviceIP")
    activity: Optional[List[str]] = None
    state_vedge_list: Optional[str] = Field(
        default=None, validation_alias="state_vedgeList", serialization_alias="state_vedgeList"
    )
    cert_install_status: Optional[str] = Field(
        default=None, validation_alias="certInstallStatus", serialization_alias="certInstallStatus"
    )
    org: Optional[str] = None
    personality: Optional[str] = None
    expiration_status: Optional[str] = Field(
        default=None, validation_alias="expirationStatus", serialization_alias="expirationStatus"
    )
    life_cycle_required: Optional[bool] = Field(
        default=None, validation_alias="lifeCycleRequired", serialization_alias="lifeCycleRequired"
    )
    hardware_cert_serial_number: Optional[str] = Field(
        default=None, validation_alias="hardwareCertSerialNumber", serialization_alias="hardwareCertSerialNumber"
    )
    subject_serial_number: Optional[str] = Field(
        default=None, validation_alias="subjectSerialNumber", serialization_alias="subjectSerialNumber"
    )
    resource_group: Optional[str] = Field(
        default=None, validation_alias="resourceGroup", serialization_alias="resourceGroup"
    )
    id: Optional[str] = None
    tags: Optional[Union[List[str], List[Dict]]] = None
    draft_mode: Optional[str] = Field(default=None, validation_alias="draftMode", serialization_alias="draftMode")
    solution: Optional[str] = None
    device_lock: Optional[str] = Field(default=None, validation_alias="device-lock", serialization_alias="device-lock")
    managed_by: Optional[str] = Field(default=None, validation_alias="managed-by", serialization_alias="managed-by")
    configured_site_id: Optional[str] = Field(
        default=None, validation_alias="configuredSiteId", serialization_alias="configuredSiteId"
    )
    ncs_device_name: Optional[str] = Field(
        default=None, validation_alias="ncsDeviceName", serialization_alias="ncsDeviceName"
    )
    config_status_message: Optional[str] = Field(
        default=None, validation_alias="configStatusMessage", serialization_alias="configStatusMessage"
    )
    template: Optional[str] = Field(default=None)
    template_id: Optional[str] = Field(default=None, validation_alias="templateId", serialization_alias="templateId")
    template_apply_log: Optional[List[str]] = Field(
        default=None, validation_alias="templateApplyLog", serialization_alias="templateApplyLog"
    )
    template_status: Optional[str] = Field(
        default=None, validation_alias="templateStatus", serialization_alias="templateStatus"
    )
    config_status_message_details: Optional[str] = Field(
        default=None, validation_alias="configStatusMessageDetails", serialization_alias="configStatusMessageDetails"
    )
    device_enterprise_certificate: Optional[str] = Field(
        default=None, validation_alias="deviceEnterpriseCertificate", serialization_alias="deviceEnterpriseCertificate"
    )
    service_personality: Optional[str] = Field(
        default=None, validation_alias="servicePersonality", serialization_alias="servicePersonality"
    )
    upload_source: Optional[str] = Field(
        default=None, validation_alias="uploadSource", serialization_alias="uploadSource"
    )
    time_remaining_for_expiration: Optional[int] = Field(
        default=None, validation_alias="timeRemainingForExpiration", serialization_alias="timeRemainingForExpiration"
    )
    domain_id: Optional[str] = Field(default=None, validation_alias="domain-id", serialization_alias="domain-id")
    local_system_ip: Optional[str] = Field(
        default=None, validation_alias="local-system-ip", serialization_alias="ocal-system-ip"
    )
    system_ip: Optional[str] = Field(default=None, validation_alias="system-ip", serialization_alias="system-ip")
    model_sku: Optional[str] = Field(default=None)
    site_id: Optional[str] = Field(default=None, validation_alias="site-id", serialization_alias="site-id")
    host_name: Optional[str] = Field(default=None, validation_alias="host-name", serialization_alias="host-name")
    sp_organization_name: Optional[str] = Field(
        default=None, validation_alias="sp-organization-name", serialization_alias="sp-organization-name"
    )
    version: Optional[str] = Field(default=None)
    vbond: Optional[str] = Field(default=None)
    vmanage_system_ip: Optional[str] = Field(
        default=None, validation_alias="vmanage-system-ip", serialization_alias="vmanage-system-ip"
    )
    vmanage_connection_state: Optional[str] = Field(
        default=None, validation_alias="vmanageConnectionState", serialization_alias="vmanageConnectionState"
    )
    last_updated: Optional[int] = Field(default=None, validation_alias="lastupdated", serialization_alias="lastupdated")
    reachability: Optional[str] = Field(default=None)
    uptime_date: Optional[int] = Field(default=None, validation_alias="uptime-date", serialization_alias="uptime-date")
    default_version: Optional[str] = Field(
        default=None, validation_alias="defaultVersion", serialization_alias="defaultVersion"
    )
    organization_name: Optional[str] = Field(
        default=None, validation_alias="organization-name", serialization_alias="organization-name"
    )
    available_versions: Optional[List[str]] = Field(
        default=None, validation_alias="availableVersions", serialization_alias="availableVersions"
    )
    site_name: Optional[str] = Field(default=None, validation_alias="site-name", serialization_alias="site-name")


class DeviceDetailsQueryParams(BaseModel):
    model: Optional[str] = None
    state: Optional[List[str]] = None
    uuid: Optional[List[str]] = None
    device_ip: Optional[List[str]] = Field(default=None, validation_alias="deviceIP", serialization_alias="deviceIP")
    validity: Optional[List[str]] = None
    family: Optional[str] = None


Validity = Literal["valid", "invalid"]


class SmartAccountSyncParams(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    password: str
    username: str
    validity_string: Validity = Field(default="valid")


class ProcessId(BaseModel):
    process_id: str = Field(validation_alias="processId", serialization_alias="processId")


class SerialFilePayload(CustomPayloadType):
    def __init__(self, image_path: str, validity: Validity = "valid"):
        self.image_path = image_path
        self.validity = validity
        self.data = open(self.image_path, "rb")
        self.fields = {"validity": self.validity, "upload": True}

    def prepared(self) -> PreparedPayload:
        return PreparedPayload(files={"file": (Path(self.data.name).name, self.data)}, data=self.fields)


ConfigType = Literal["cloudinit", "encodedstring"]


class GenerateBoostrapConfigurationQueryParams(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    configtype: Optional[ConfigType] = Field(default="cloudinit")
    incl_def_root_cert: Optional[bool] = Field(
        default=False, validation_alias="inclDefRootCert", serialization_alias="inclDefRootCert"
    )
    version: Optional[str] = Field(default="v1")


class BoostrapConfiguration(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    bootstrap_config: Optional[str] = Field(
        default=None, validation_alias="bootstrapConfig", serialization_alias="bootstrapConfig"
    )


class UploadSerialFileResponse(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    vedge_list_upload_msg: Optional[str] = Field(
        default=None, validation_alias="vedgeListUploadMsg", serialization_alias="vedgeListUploadMsg"
    )
    vedge_list_upload_status: Optional[str] = Field(
        default=None, validation_alias="vedgeListUploadStatus", serialization_alias="vedgeListUploadStatus"
    )
    id: Optional[str] = None
    vedge_list_status_code: Optional[str] = Field(
        default=None, validation_alias="vedgeListStatusCode", serialization_alias="vedgeListStatusCode"
    )
    activity_list: Optional[Union[List, str]] = Field(
        default=None, validation_alias="activityList", serialization_alias="activityList"
    )


class ConfigurationDeviceInventory(APIEndpoints):
    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/system/device/{device_uuid}/unlock")
    def unlock(self, device_uuid: UUID, payload: DeviceUnlockPayload) -> DeviceUnlockResponse:
        ...

    @post("/system/device")
    def create_device(self, payload: DeviceCreationPayload) -> None:
        ...

    @delete("/system/device/{uuid}")
    def delete_device(self, uuid: UUID) -> DeviceDeletionResponse:
        ...

    # Covers:
    # url = "/dataservice/system/device/controllers"
    # url = "/dataservice/system/device/vedges"
    @get("/system/device/{device_category}", "data")
    def get_device_details(
        self, device_category: DeviceCategory, params: DeviceDetailsQueryParams = DeviceDetailsQueryParams()
    ) -> DataSequence[DeviceDetailsResponse]:
        ...

    @post("/system/device/smartaccount/sync")
    def sync_devices_from_smart_account(self, payload: SmartAccountSyncParams) -> ProcessId:
        ...

    @post("/system/device/fileupload")
    def upload_wan_edge_list(self, payload: SerialFilePayload) -> UploadSerialFileResponse:
        ...

    @get("/system/device/bootstrap/device/{uuid}")
    def generate_bootstrap_configuration(
        self, uuid: UUID, params: GenerateBoostrapConfigurationQueryParams = GenerateBoostrapConfigurationQueryParams()
    ) -> BoostrapConfiguration:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Literal, Union

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable, _ParcelBase

ConfigTypeValue = Literal["non-eSim"]


class ControllerConfig(BaseModel):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True, populate_by_name=True)
    id: Union[Variable, Global[str]] = Field(min_length=1, max_length=5, description="Cellular ID")
    slot: Union[Variable, Global[int], Default[int], None] = Field(
        default=None, description="Set primary SIM slot", ge=0, le=1
    )
    max_retry: Union[Variable, Global[int], Default[None], None] = Field(
        default=None,
        description="Set SIM failover retries",
        ge=0,
        le=65535,
        serialization_alias="maxRetry",
        validation_alias="maxRetry",
    )
    failover_timer: Union[Variable, Global[int], Default[None], None] = Field(
        default=None,
        description="Set SIM failover timeout in minutes",
        ge=3,
        le=7,
        serialization_alias="failovertimer",
        validation_alias="failovertimer",
    )
    auto_sim: Union[Variable, Global[bool], Default[None], None] = Field(
        default=None,
        description="Enable/Disable Firmware Auto Sim",
        serialization_alias="autoSim",
        validation_alias="autoSim",
    )


class CellularController(_ParcelBase):
    config_type: Default[ConfigTypeValue] = Field(
        default=Default(value="non-eSim"), validation_alias=AliasPath("data", "configType")
    )
    controller_config: ControllerConfig = Field(validation_alias=AliasPath("data", "controllerConfig"))

    @staticmethod
    def add_controller_config(
        id: Union[Variable, Global[str]],
        max_retry: Union[Variable, Global[int], Default[None], None] = None,
        failover_timer: Union[Variable, Global[int], Default[None], None] = None,
        auto_sim: Union[Variable, Global[bool], Default[None], None] = None,
    ):
        return ControllerConfig(id=id, max_retry=max_retry, failover_timer=failover_timer, auto_sim=auto_sim)
